/* $Workfile: T2mDemo.c $       $Revision: 103 $      $Modtime: 9/12/03 11:09a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2MDEMO.C

    A simple example console application demonstrating the use of OTX Driver
    with the following boards:

      Thor-2-PCMCIA series:

      Thor-2-PCMCIA, 
      Thor-2-PCMCIA-Plus, 
      Thor-2-PCMCIA-Ex, 
      Thor-2-PCMCIA-Pro,
      Thor-2-PCMCIA-CST

    or 

      Thor-2-PCI / Thor-8-PCI + Vidar-5x4-ASM,
      Thor-2-PCI / Thor-8-PCI + Vidar-5x4-ASM-CST,
      Thor-2-PCI / Thor-8-PCI + Vidar-5x4-ASM-PRO,
      Thor-2-PCI / Thor-8-PCI + Vidar-5x4-ASM-EX,
      Thor-2-PCI / Thor-8-PCI + Vidar-55x4-ASM,


    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 E1 Lines and expects them to be looped with an external
    loop cable. The LI Internal HDLC controller is used to send HDLC frames from
    one Line Interface to be received by the other.

    A DTMF detector and DTMF dialler are created as logical devices in the two DSPs.
    A Data Converter device is used to create an internal cross-connect with the DSP .

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(_WIN32_WCE)
#include "Lib/WindowsCe/OsaWindowsCe.h"
#else   // #if defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxBrd.h"
#include "DspIoctl.h"

#define DEMO_DEFAULT_LOG_FILE "T2mDemo.log"
#define DEMO_DEFAULT_BOARD_NO 0

#define DEMO_TASK_REF 100L               // Task Reference for a non-blocking call

//#define DEMO_SUPER_CHANNEL 1  // Uncomment this line to use a 256 kbits/s channel on the E1T1 HDLC Sender/Receiver

//#define DEMO_USE_TIMESTAMP_FROM_1601 1  // Uncomment this line to use timestamps with a start time of Jan 1, 1601

#if defined(_WIN32_WCE)
#define DEMO_DIALER_MAX 8
#else
#define DEMO_DIALER_MAX 64
#endif

#if defined(__linux__)
#define MAX_PATH 260    // linux doesn't define MAX_PATH
#endif

typedef struct DemoEnvT {
  char               *pszInputStr;
  OTX_BOOL            bExitOnError;
  char                cwd[MAX_PATH];
  char                logFileName[MAX_PATH];
  enum OtxPhysicalDeviceTypeE nBoardType;
  int                 coLi;
  int                 coDsp;
  OtxDataFormatE      law;
  int                 boardNo;
  OtxT1E1LiModeE      eT1E1Mode;    // T1 or E1 mode
  OtxBrdClockSourceE  eClkSource;
  OtxDspSpmE          eProgramModule;      // Standard Program Module to be loaded in the DSP (from the Lib)
  OTX_BOOL            bUseDspDebugger;
} DemoEnvT;



/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);
static void DemoCrossConnect(OTX_INT32 nTsA, OTX_INT32 nTsB);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);

/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
OTX_INT32   g_cbLiHdlcSenderFrame  = 5;
OTX_INT32   g_cbDspHdlcSenderFrame = 5;
OTX_INT32   g_cbHdlcFrame = 10;
OTX_INT32   g_coHdlcSenderFrames   = 1;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bHdlcVerbose = OTX_TRUE;
OTX_UINT32  g_coHdlcRxFrames[256];
OTX_UINT32  g_coHdlcBadRxFrames[256];
OTX_UINT8   g_sqHdlcFrame[OD_HDLC_MAX_TX_HDLC_FRAME_LEN];
OTX_UINT32  g_nFrequency;
OTX_BOOL    g_bAutoTransmit[2];

static OsaMutexS g_analyzeCritSec;
static OsaMutexS g_demoLogfCritSec;


struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2]; // Only used for Thor-8-PCI)
  OTX_HANDLE hVidarAsmController0;  // Only used for Thor-2-PCI and Thor-8-PCI (not for Thor-2-PCMCIA)
  OTX_HANDLE hLi[2];
  OTX_HANDLE hDsp[4];
} g_hPhysDevs;



struct {
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;
} g_sqLiHdlcSender[2];

struct {
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;
} g_sqLiHdlcReceiver[2];

struct {
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;
} g_sqDspHdlcSender[8];

struct {
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;
} g_sqDspHdlcReceiver[8];


struct {
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;
} g_sqDtmfDetector[2];

struct {
  // Use the same pipe as the g_sqDtmfDetector pipe
  OTX_HANDLE hHandle;
} g_sqSilenceDetector[2];

struct {
  OTX_HANDLE hPipe;  // This pipe is the output from Data Converter. We share an DtmfDetector 
                     // pipe as the input pipe to this device.
  OTX_HANDLE hHandle;

  OTX_INT32 nToTs;
} g_sqDataConverter[2];

struct {
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;
} g_sqConstValGenerator[64];

struct {
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;
} g_sqValSampler[64];


struct {
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;

  OTX_INT32 nDialer;

  OTX_EVENT   hNotificationEvent;
  OTX_HANDLE  hEventQueue;

  char sqDigits[16];

  OTX_INT32 coDtmfCompleted;
  OTX_INT32 nRepeatTimes;

  OTX_BOOL  bAbort;
  OTX_BOOL  bActive;

} g_sqDtmfDialer[64];

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;



/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(/*128*/2048, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hEventQueue, result);

}



void DemoCreatePhysicalDevices(void)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the Thor-2 board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.


  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening OTX_DEVICE_THOR_PCI board\n");
    break;
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening OTX_DEVICE_THOR_PCMCIA board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

#if 0
  // Disable DSP Serial port clocks (so that we can enable them synchronously after the DSP programs are loaded)
  result = OtxDrvSetAttributeValueBool(
    g_hPhysDevs.hThorController0, 
    OTX_ATTR_BRD_DSP_PORTS_DISABLED, 
    OTX_TRUE
    ); 
  DemoCheckRc(g_hPhysDevs.hThorController0, result);  
#endif
  
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_THOR_PCI:
    // On the Thor-2-PCI and the Thor-2-PCMCIA board the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < 2; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_8_PCI:
    {
      OTX_INT32 nQuadE1T1; // Only used for Thor-8-PCI

      // On the Thor-8-PCI board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < 2; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  


  // Open physical DSP devices
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
    // Get handles to the physcial DSP devices #0 and #1
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_8_PCI:
    {
      OTX_UINT32   eDevType;
      OTX_ITERATOR itrChildTypes;
      OtxDrvIteratorBegin(&itrChildTypes);
      while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hThorController0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
        OTX_BOOL bOpenAsm = OTX_FALSE;
        switch (eDevType) {
        case OTX_DEVICE_VIDAR_5x_ASM:
          DemoLogf("Opening Vidar-5x4-ASM board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_CST:
          DemoLogf("Opening Vidar-5x4-ASM-CST board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_V90:
          DemoLogf("Opening Vidar-5x4-ASM-V90 board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_EX:
          DemoLogf("Opening Vidar-5x4-ASM-EX board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_PRO:
          DemoLogf("Opening Vidar-5x4-ASM-PRO board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_55x4_ASM:
          DemoLogf("Opening Vidar-55x4-ASM board\n");
          bOpenAsm = OTX_TRUE;

        // Verify that the system supports the Vidar-55x4-ASM board
        {
          char szCpldBuildNumbers[256];
          char szBrdBuildNumber[16];
          OTX_INT32 nBoardBuildNo;
          
          if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCI) {
            
            //  Verify firmware revision
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            if (strstr(szCpldBuildNumbers, "OTXCDC: 0761") || strstr(szCpldBuildNumbers, "OTXCDC: 0762")) {
              DemoLogf("Error: The firmware on the Thor-2-PCI must be upgraded to firmware build 5\n");
              DemoLogf("       (OTXCDC 0764) or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
              DemoLogf("       web site (www.odints.com)\n");
              DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
              DemoExitQuestion();
            }
            
            // Verify board build number
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            nBoardBuildNo = strtol(strtok(szBrdBuildNumber, "."), NULL, 0);
            DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
            if (nBoardBuildNo < 11) {
              DemoLogf("Error: The Thor-2-PCI board must be upgraded to build 11\n");
              DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
              DemoExitQuestion();
            }
            
          } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) {
            //  Verify firmware revision
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            if (strstr(szCpldBuildNumbers, "OTXCDC: 0771")) {
              DemoLogf("Error: The firmware on the Thor-8-PCI must be upgraded to firmware build 3\n");
              DemoLogf("       (OTXCDC 0772) or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
              DemoLogf("       web site (www.odints.com)\n");
              DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
              DemoExitQuestion();
            }
            
            // Verify board build number
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            nBoardBuildNo = strtol(szBrdBuildNumber, NULL, 0);
            DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
            if (nBoardBuildNo < 5) {
              DemoLogf("Error: The Thor-8-PCI board must be upgraded to build 5\n");
              DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
              DemoExitQuestion();
            }
            
          }
          
        }
          break;
        }
        if (bOpenAsm) {
          // Get a handle to to the ASM Daughter Board
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, (OtxPhysicalDeviceTypeE) eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
        }

      }
      
      for (i = 0; i < g_demoEnv.coDsp; i++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
      }
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  
  
  // Load DSP Program
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    switch (g_demoEnv.eProgramModule) {
    case OTX_DSP_SPM_OTXSPM1:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM1\n", i);
      break;
    case OTX_DSP_SPM_OTXSPM2:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2\n", i);
      break;
    case OTX_DSP_SPM_OTXHDLC:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXHDLC\n", i);
      break;
    case OTX_DSP_SPM_OTXDTTX:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXDTTX\n", i);
      break;
    default:
      DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }

    result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], g_demoEnv.eProgramModule); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

  if (g_demoEnv.bUseDspDebugger) {
    DemoLogf("Please load and run DSP to DSP#0 and DSP#1 in your DSP Debugger\n");
    DemoLogf("Press any key when the program is running...\n");
    OsaConsoleGetCh();
  }

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));
  
  for (i = 0; i < 2; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    // Configure Transceivers
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);

  }

}



void DemoEnablePhysicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Enable the board (Physical Device)
  DemoLogf("Enable hThorController0\n");
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  if (g_hPhysDevs.hVidarAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
  }

  // If board type is Thor-8-PCI then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) {
    OTX_INT32 nQuadE1T1; // Only used for Thor-8-PCI

    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }


  for (i = 0; i < 2; i++) {

    // Enable Physical Devices
    DemoLogf("Enable g_hPhysDevs.hLi[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    DemoLogf("g_hPhysDevs.hDsp[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);

  }

}


void DemoCreateLogicalDevices(
/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

    Logical Devices can be classified into Control Device and Data Devices:

    - Control Device:  A logical device performing a control function. E.g. Event Queue.

    - Data Filter Device: A device implemented in Hardware, Firmware of Software
    which converts, processes, or manipulates serial data. Examples of data devices include
    Time-Space Switch, HDLC Controller, Tone Detector, Tone Generator, etc.

    Data Filter devices can be categorized into 3 basic types:

     - Data Source: Originates a Pipe or adds data to an existing Pipe. E.g. DTMF Tone Generator.

     - Data Sink: Terminates a Pipe or extracts data from a Pipe. E.g. DTMF Detector.

     - Data Bridge: Cross-connects data from one Pipe to another or combines data from
     multiple Pipes into one Pipe. E.g. cross-connect in TSS.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;

  // LI#0 and LI#1: Create Hdlc Senders and Receivers logical devices hosted in the physical Line Interface devices
  for (i = 0; i < 2; i++) {
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hLi[i], OTX_LDEVICE_HDLC_SENDER,   g_hEventQueue, i+100, &g_sqLiHdlcSender[i].hHandle); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hLi[i], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i+100, &g_sqLiHdlcReceiver[i].hHandle); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }


  switch (g_demoEnv.eProgramModule) {
  case OTX_DSP_SPM_OTXSPM1:
  case OTX_DSP_SPM_OTXSPM2:
    for (i = 0; i < 2; i++) {
      // Create Hdlc Senders and Receivers logical devices hosted in the physical DSP devices
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_HDLC_SENDER,   g_hEventQueue, 2+i, &g_sqDspHdlcSender[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, 2+i, &g_sqDspHdlcReceiver[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
    for (i = 0; i < 2; i++) {
      // Create Dtmf generator
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_DTMF_DIALER, g_hEventQueue, i, &g_sqDtmfDialer[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }  
    for (i = 0; i < 2; i++) {
      // Create Dtmf Detector
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_DTMF_DETECTOR, g_hEventQueue, i, &g_sqDtmfDetector[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
    for (i = 0; i < 2; i++) {
      // Create Silence Detector
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_SILENCE_DETECTOR, g_hEventQueue, i, &g_sqSilenceDetector[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
    for (i = 0; i < 2; i++) {
      // Create Data Converter (for cross-connection with the DSP)
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_DATA_CONVERTER, g_hEventQueue, i, &g_sqDataConverter[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
    break;
  case OTX_DSP_SPM_OTXDTTX:
    // Create 64 DTMF Dialers (32 on DSP#0 and 32 on DSP#1)
    for (i = 0; i < 2; i++) {
      OTX_INT32 j;
      for (j = 0; j < 32; j++) {
        if ((i*32 + j)%16 == 0) continue; // Skip 0 (sync), 16 (signalling), 32 (sync), 48 (signalling)

        DemoLogf("Creating logical device [%d]\n", i*32 + j);

        // Create a notification Event, Event Queue for each DTMF Dialer
        result = OtxDrvCreateEvent(&g_sqDtmfDialer[i*32 + j].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
        result = OtxDrvCreateEventQueue(32, g_sqDtmfDialer[i*32 + j].hNotificationEvent, &g_sqDtmfDialer[i*32 + j].hEventQueue); DemoCheckRc(g_sqDtmfDialer[i*32 + j].hEventQueue, result);
        result = OtxDrvEnable(g_sqDtmfDialer[i*32 + j].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDialer[i*32 + j].hEventQueue, result);

        // Create Dtmf generator (assign it to its own Event Queue)
        result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_DTMF_DIALER, g_sqDtmfDialer[i*32 + j].hEventQueue, i*32 + j, &g_sqDtmfDialer[i*32 + j].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);

        // Create Const Val Generator
        result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_CONST_VAL_GENERATOR, g_hEventQueue, i*32 + j, &g_sqConstValGenerator[i*32 + j].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
        // Create Val Sampler
        result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_VAL_SAMPLER, g_hEventQueue, i*32 + j, &g_sqValSampler[i*32 + j].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      }
    }  
    break;
  case OTX_DSP_SPM_OTXHDLC:
    for (i = 0; i < 8; i++) {
      // Create Hdlc Senders and Receivers logical devices hosted in the physical DSP devices
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i/4], OTX_LDEVICE_HDLC_SENDER,   g_hEventQueue, i, &g_sqDspHdlcSender[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i/4], result);
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i/4], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i, &g_sqDspHdlcReceiver[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i/4], result);
    }
    break;    
  default:
    DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
}


}


void DemoCloseLogicalDevices(
/*******************************************************************************

  @func
    Close Logical Devices

*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;

  // LI#0 and LI#1: Create Hdlc Senders and Receivers logical devices hosted in the physical Line Interface devices
  for (i = 0; i < 2; i++) {
    result = OtxDrvClose(g_sqLiHdlcSender[i].hHandle); DemoCheckRc(g_sqLiHdlcSender[i].hHandle, result);
    result = OtxDrvClose(g_sqLiHdlcReceiver[i].hHandle); DemoCheckRc(g_sqLiHdlcReceiver[i].hHandle, result);
  }


  switch (g_demoEnv.eProgramModule) {
  case OTX_DSP_SPM_OTXSPM1:
  case OTX_DSP_SPM_OTXSPM2:
    for (i = 0; i < 2; i++) {
      // Create Hdlc Senders and Receivers logical devices hosted in the physical DSP devices
      result = OtxDrvClose(g_sqDspHdlcSender[i].hHandle); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
      result = OtxDrvClose(g_sqDspHdlcReceiver[i].hHandle); DemoCheckRc(g_sqDspHdlcReceiver[i].hHandle, result);
    }
    for (i = 0; i < 2; i++) {
      // Create Dtmf generator
      result = OtxDrvClose(g_sqDtmfDialer[i].hHandle); DemoCheckRc(g_sqDtmfDialer[i].hHandle, result);
    }  
    for (i = 0; i < 2; i++) {
      // Create Dtmf Detector
      result = OtxDrvClose(g_sqDtmfDetector[i].hHandle); DemoCheckRc(g_sqDtmfDetector[i].hHandle, result);
    }
    for (i = 0; i < 2; i++) {
      // Create Silence Detector
      result = OtxDrvClose(g_sqSilenceDetector[i].hHandle); DemoCheckRc(g_sqSilenceDetector[i].hHandle, result);
    }
    break;
  case OTX_DSP_SPM_OTXDTTX:
    // Create 64 DTMF Dialers (32 on DSP#0 and 32 on DSP#1)
    for (i = 0; i < 2; i++) {
      OTX_INT32 j;
      for (j = 0; j < 32; j++) {
        if ((i*32 + j)%16 == 0) continue; // Skip 0 (sync), 16 (signalling), 32 (sync), 48 (signalling)

        DemoLogf("Creating logical device [%d]\n", i*32 + j);

        // Create a notification Event, Event Queue for each DTMF Dialer
        //result = OtxDrvCreateEvent(&g_sqDtmfDialer[i*32 + j].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
        //result = OtxDrvCreateEventQueue(32, g_sqDtmfDialer[i*32 + j].hNotificationEvent, &g_sqDtmfDialer[i*32 + j].hEventQueue); DemoCheckRc(g_sqDtmfDialer[i*32 + j].hEventQueue, result);
        //result = OtxDrvEnable(g_sqDtmfDialer[i*32 + j].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDialer[i*32 + j].hEventQueue, result);

        // Create Dtmf generator (assign it to its own Event Queue)
        result = OtxDrvClose(g_sqDtmfDialer[i*32 + j].hHandle); DemoCheckRc(g_sqDtmfDialer[i*32 + j].hHandle, result);

        // Create Const Val Generator
        result = OtxDrvClose(g_sqConstValGenerator[i*32 + j].hHandle); DemoCheckRc(g_sqConstValGenerator[i*32 + j].hHandle, result);
        // Create Val Sampler
        result = OtxDrvClose(g_sqValSampler[i*32 + j].hHandle); DemoCheckRc(g_sqValSampler[i*32 + j].hHandle, result);
      }
    }  
    break;
  case OTX_DSP_SPM_OTXHDLC:
    for (i = 0; i < 2; i++) {
      // Create Hdlc Senders and Receivers logical devices hosted in the physical DSP devices
      result = OtxDrvClose(g_sqDspHdlcSender[i].hHandle); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
      result = OtxDrvClose(g_sqDspHdlcReceiver[i].hHandle); DemoCheckRc(g_sqDspHdlcReceiver[i].hHandle, result);
    }
    break;    
  default:
    DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
  }


}


void DemoConnectDevices(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
  void
)
{
  OTX_INT32 ts;
  OTX_INT32 nCapacity;
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];

  // Connect the HDLC Sender/Receiver Logical Device in the Line Interface (one for each direction)
  for (i = 0; i < 2; i++) {
    OTX_INT32 nCapacity;
    {
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        // T1 mode
        ts = 24; 
      } else {
        // E1 mode
        ts = 16; 
      }

      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));

#ifdef DEMO_SUPER_CHANNEL
      //memset(sqTimeSlotMask+24, 0xFF, 8 * sizeof(OTX_UINT8));
      //nCapacity = 8*64;
      memset(sqTimeSlotMask+ts-4, 0xFF, 4 * sizeof(OTX_UINT8));
      nCapacity = 4*64;
#else
      sqTimeSlotMask[ts] = 0xFF;
      nCapacity = 64;
#endif      
      
      // Create an incoming and outgoing Pipe representing the external T1/E1
      result = OtxDrvCreatePipe(nCapacity, &g_sqLiHdlcReceiver[i].hPipe); DemoCheckRc(g_sqLiHdlcReceiver[i].hPipe, result);
      result = OtxDrvCreatePipe(nCapacity, &g_sqLiHdlcSender[i].hPipe); DemoCheckRc(g_sqLiHdlcSender[i].hPipe, result);
      
      // Connect external T1/E1 pipes. Only connect one end of each pipe. External Interface is LI Highway Number 1
      result = OtxDrvConnectPipe(g_sqLiHdlcReceiver[i].hPipe,  OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_sqLiHdlcReceiver[i].hPipe, result);
      result = OtxDrvConnectPipe(g_sqLiHdlcSender[i].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_sqLiHdlcSender[i].hPipe, result);

      // Connect the logical devices within the Li
      result = OtxDrvConnectLogicalDevice(g_sqLiHdlcSender[i].hHandle, -1, g_sqLiHdlcSender[i].hPipe, -1); DemoCheckRc(g_sqLiHdlcSender[i].hHandle, result);
      result = OtxDrvConnectLogicalDevice(g_sqLiHdlcReceiver[i].hPipe,           -1, g_sqLiHdlcReceiver[i].hHandle, -1); DemoCheckRc(g_sqLiHdlcReceiver[i].hPipe, result);
    }
  }




  // CONNECT LOGICAL DEVICES WITHIN THE DSP
  // --------------------------------------
 for (i = 0; i < 2; i++) {

   switch (g_demoEnv.eProgramModule) {
   case OTX_DSP_SPM_OTXSPM1:
   case OTX_DSP_SPM_OTXSPM2:
     // Receive direction
     {
       DemoLogf("Connect DSP HDLC Receiver\n");
       ts = 1;
       memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
       sqTimeSlotMask[ts] = 0xFF;
       nCapacity = 64;
       
       result = OtxDrvCreatePipe(nCapacity, &g_sqDspHdlcReceiver[i].hPipe); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
       result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_OUTPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
       result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
       
       result = OtxDrvConnectLogicalDevice(g_sqDspHdlcReceiver[i].hPipe,  -1,  g_sqDspHdlcReceiver[i].hHandle,    -1); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
     }
     
     {
       DemoLogf("Connect DSP DTMF and Silence Detector\n");
       ts = 20;
       memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
       sqTimeSlotMask[ts] = 0xFF;
       nCapacity = 64;
       
       result = OtxDrvCreatePipe(nCapacity, &g_sqDtmfDetector[i].hPipe); DemoCheckRc(g_sqDtmfDetector[i].hPipe, result);
       result = OtxDrvConnectPipe(g_sqDtmfDetector[i].hPipe, OTX_PIPE_OUTPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDtmfDetector[i].hPipe, result);
       result = OtxDrvConnectPipe(g_sqDtmfDetector[i].hPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDtmfDetector[i].hPipe, result);
       
       result = OtxDrvConnectLogicalDevice(g_sqDtmfDetector[i].hPipe,  -1,  g_sqDtmfDetector[i].hHandle,    -1); DemoCheckRc(g_sqDtmfDetector[i].hPipe, result);

       // Also connect the Silence detector to the same pipe
       result = OtxDrvConnectLogicalDevice(g_sqDtmfDetector[i].hPipe,  -1,  g_sqSilenceDetector[i].hHandle,    -1); DemoCheckRc(g_sqDtmfDetector[i].hPipe, result);
     }
          
     // Transmit direction
     {
       DemoLogf("Connect DSP DTMF Dialer\n");
       ts = 20;
       memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
       sqTimeSlotMask[ts] = 0xFF;
       nCapacity = 64;
       
       result = OtxDrvCreatePipe(nCapacity, &g_sqDtmfDialer[i].hPipe); DemoCheckRc(g_sqDtmfDialer[i].hPipe, result);
       result = OtxDrvConnectPipe(g_sqDtmfDialer[i].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDtmfDialer[i].hPipe, result);
       result = OtxDrvConnectPipe(g_sqDtmfDialer[i].hPipe, OTX_PIPE_OUTPUT,   g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDtmfDialer[i].hPipe, result);
       
       result = OtxDrvConnectLogicalDevice(g_sqDtmfDialer[i].hHandle,  -1, g_sqDtmfDialer[i].hPipe,  -1); DemoCheckRc(g_sqDtmfDialer[i].hHandle, result);
     }
          
     {
       DemoLogf("Connect DSP HDLC Sender\n");
       ts = 1;
       memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
       sqTimeSlotMask[ts] = 0xFF;
       nCapacity = 64;
       
       result = OtxDrvCreatePipe(nCapacity, &g_sqDspHdlcSender[i].hPipe); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
       result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
       result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe, OTX_PIPE_OUTPUT,   g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
       
       result = OtxDrvConnectLogicalDevice(g_sqDspHdlcSender[i].hHandle,  -1, g_sqDspHdlcSender[i].hPipe,  -1); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
     }
     // Bidirectional devices
     {
       DemoLogf("Connect DSP Data Converter (cross-connection within the DSP)\n");
       
       // Use the DtmfDetector pipe as the input to the data converter.
       // I.e. this pipe will now feed data to both the DTMF detector and the Data Converter devices
       result = OtxDrvConnectLogicalDevice(g_sqDtmfDetector[i].hPipe,  -1,  g_sqDataConverter[i].hHandle,  -1); DemoCheckRc(g_sqDataConverter[i].hPipe, result);

       // Output from Data Converter
       ts = 21;
       g_sqDataConverter[i].nToTs = ts;
       memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
       sqTimeSlotMask[ts] = 0xFF;
       nCapacity = 64;
       
       result = OtxDrvCreatePipe(nCapacity, &g_sqDataConverter[i].hPipe); DemoCheckRc(g_sqDataConverter[i].hPipe, result);
       result = OtxDrvConnectPipe(g_sqDataConverter[i].hPipe, OTX_PIPE_INPUT,    g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDataConverter[i].hPipe, result);
       result = OtxDrvConnectPipe(g_sqDataConverter[i].hPipe, OTX_PIPE_OUTPUT,   g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDataConverter[i].hPipe, result);
       
       result = OtxDrvConnectLogicalDevice(g_sqDataConverter[i].hHandle,  -1, g_sqDataConverter[i].hPipe,  -1); DemoCheckRc(g_sqDataConverter[i].hHandle, result);

     }

     break;

   case OTX_DSP_SPM_OTXDTTX:
     {
       OTX_INT32 j;
       for (j = 0; j < 32; j++) {
         if ((i*32 + j)%16 == 0) continue; // Skip 0 (sync), 16 (signalling), 32 (sync), 48 (signalling)

         DemoLogf("Connecting Logical device [%d]\n", i*32 + j);

         // Receive direction:
         {
           //DemoLogf("Connect DSP Val Sampler\n");
           ts = j;
           memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
           sqTimeSlotMask[ts] = 0xFF;
           nCapacity = 64;
           
           result = OtxDrvCreatePipe(nCapacity, &g_sqValSampler[i*32 + j].hPipe); DemoCheckRc(g_sqValSampler[i*32 + j].hPipe, result);
           result = OtxDrvConnectPipe(g_sqValSampler[i*32 + j].hPipe, OTX_PIPE_OUTPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqValSampler[i*32 + j].hPipe, result);
           result = OtxDrvConnectPipe(g_sqValSampler[i*32 + j].hPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqValSampler[i*32 + j].hPipe, result);
           
           result = OtxDrvConnectLogicalDevice(g_sqValSampler[i*32 + j].hPipe, -1, g_sqValSampler[i*32 + j].hHandle, -1); DemoCheckRc(g_sqValSampler[i*32 + j].hPipe, result);
         }

         // Transmit direction
         {
           //DemoLogf("Connect DSP DTMF Dialer and ConstVal Generator\n");
           ts = j;
           memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
           sqTimeSlotMask[ts] = 0xFF;
           nCapacity = 64;
           
           result = OtxDrvCreatePipe(nCapacity, &g_sqDtmfDialer[i*32 + j].hPipe); DemoCheckRc(g_sqDtmfDialer[i*32 + j].hPipe, result);

           // The DtmfDialer and the ConstValGenetor logical devices share the same pipe
           g_sqConstValGenerator[i*32 + j].hPipe = g_sqDtmfDialer[i*32 + j].hPipe;

           result = OtxDrvConnectPipe(g_sqDtmfDialer[i*32 + j].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDtmfDialer[i*32 + j].hPipe, result);
           result = OtxDrvConnectPipe(g_sqDtmfDialer[i*32 + j].hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDtmfDialer[i*32 + j].hPipe, result);
           
           result = OtxDrvConnectLogicalDevice(g_sqDtmfDialer[i*32 + j].hHandle,  -1, g_sqDtmfDialer[i*32 + j].hPipe,  -1); DemoCheckRc(g_sqDtmfDialer[i*32 + j].hHandle, result);
           result = OtxDrvConnectLogicalDevice(g_sqConstValGenerator[i*32 + j].hHandle,  -1, g_sqConstValGenerator[i*32 + j].hPipe,  -1); DemoCheckRc(g_sqConstValGenerator[i*32 + j].hHandle, result);
         }

      }
     }
     break;
   case OTX_DSP_SPM_OTXHDLC:
     // Receive direction:
     {
       OTX_INT32 j;

       DemoLogf("Connect 4 DSP HDLC Receivers (4*16 kpbs pipe on TS1)\n");
       ts = 1;
       memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
       sqTimeSlotMask[ts] = 0x03;
       nCapacity = 16;
       
       for (j = 0; j < 4; j++) {
         result = OtxDrvCreatePipe(nCapacity, &g_sqDspHdlcReceiver[i*4 + j].hPipe); DemoCheckRc(g_sqDspHdlcReceiver[i*4 + j].hPipe, result);
         result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i*4 + j].hPipe, OTX_PIPE_OUTPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcReceiver[i*4 + j].hPipe, result);
         result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i*4 + j].hPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcReceiver[i*4 + j].hPipe, result);
         
         result = OtxDrvConnectLogicalDevice(g_sqDspHdlcReceiver[i*4 + j].hPipe,  -1,  g_sqDspHdlcReceiver[i*4 + j].hHandle,    -1); DemoCheckRc(g_sqDspHdlcReceiver[i*4 + j].hPipe, result);
         
         sqTimeSlotMask[ts] = sqTimeSlotMask[ts] << 2; // Shift to the next subchannel within this timeslot

       }
     }
     
     // Transmit direction
     {
       OTX_INT32 j;

       DemoLogf("Connect 4 DSP HDLC Senders (4*16 kpbs pipe on TS1)\n");
       ts = 1;
       memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
       sqTimeSlotMask[ts] = 0x03;
       nCapacity = 16;
       
       for (j = 0; j < 4; j++) {
         result = OtxDrvCreatePipe(nCapacity, &g_sqDspHdlcSender[i*4 + j].hPipe); DemoCheckRc(g_sqDspHdlcSender[i*4 + j].hPipe, result);
         result = OtxDrvConnectPipe(g_sqDspHdlcSender[i*4 + j].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcSender[i*4 + j].hPipe, result);
         result = OtxDrvConnectPipe(g_sqDspHdlcSender[i*4 + j].hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcSender[i*4 + j].hPipe, result);
         
         result = OtxDrvConnectLogicalDevice(g_sqDspHdlcSender[i*4 + j].hHandle,  -1, g_sqDspHdlcSender[i*4 + j].hPipe,  -1); DemoCheckRc(g_sqDspHdlcSender[i*4 + j].hHandle, result);
         
         sqTimeSlotMask[ts] = sqTimeSlotMask[ts] << 2; // Shift to the next subchannel within this timeslot
         
       }
     }
     break;    
   default:
     DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
     OsaConsoleExit();
     OtxDrvReleaseLib();
     exit(1);
   }
   
    
 }

 
}


void DemoInitializeLogicalDevices(
/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after 
    the devices are created and connected.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < 2; i++) {

    // Initialize the Hdlc Sender and Receiver in the Lis
    result = OtxHdlcSenderInit(
      g_sqLiHdlcSender[i].hHandle, // Handle to logical HDLC Filter Device
      OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
      2,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      OTX_FALSE,                 // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                  // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
      OTX_HDLC_TR_NA             // HDLC Transmission Rate for V.110 and X.30 modes
      ); DemoCheckRc(g_sqLiHdlcSender[i].hHandle, result);          
    
    result = OtxHdlcReceiverInit(
      g_sqLiHdlcReceiver[i].hHandle, // Handle to logical HDLC Filter Device
      OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
      2,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      OTX_FALSE,                 // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                  // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
      OTX_HDLC_TR_NA,            // HDLC Transmission Rate for V.110 and X.30 modes
      OTX_TRUE                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
      ); DemoCheckRc(g_sqLiHdlcReceiver[i].hHandle, result);          

    switch (g_demoEnv.eProgramModule) {
    case OTX_DSP_SPM_OTXSPM1:
    case OTX_DSP_SPM_OTXSPM2:
      // Initialize the Hdlc Sender and Receiver in the Dsps
      result = OtxHdlcSenderInit(
        g_sqDspHdlcSender[i].hHandle, // Handle to logical HDLC Filter Device
        OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
        2,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
        OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
        OTX_FALSE,                 // If OTX_TRUE then all bits are inverted when transmitted and before received
        OTX_TRUE,                  // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
        OTX_HDLC_TR_NA             // HDLC Transmission Rate for V.110 and X.30 modes
        ); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);          
      
      result = OtxHdlcReceiverInit(
        g_sqDspHdlcReceiver[i].hHandle, // Handle to logical HDLC Filter Device
        OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
        2,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
        OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
        OTX_FALSE,                 // If OTX_TRUE then all bits are inverted when transmitted and before received
        OTX_TRUE,                  // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
        OTX_HDLC_TR_NA,            // HDLC Transmission Rate for V.110 and X.30 modes
        OTX_TRUE                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
        ); DemoCheckRc(g_sqDspHdlcReceiver[i].hHandle, result);          
      break;
    case OTX_DSP_SPM_OTXHDLC:
      {
        OTX_INT32 j;
        // Initialize the Hdlc Sender and Receiver in the Dsps
        for (j = 0; j < 4; j++) {
          result = OtxHdlcSenderInit(
            g_sqDspHdlcSender[i*4 + j].hHandle, // Handle to logical HDLC Filter Device
            OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
            2,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
            OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
            OTX_FALSE,                 // If OTX_TRUE then all bits are inverted when transmitted and before received
            OTX_TRUE,                  // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
            OTX_HDLC_TR_NA             // HDLC Transmission Rate for V.110 and X.30 modes
            ); DemoCheckRc(g_sqDspHdlcSender[i*4 + j].hHandle, result);          
          
          result = OtxHdlcReceiverInit(
            g_sqDspHdlcReceiver[i*4 + j].hHandle, // Handle to logical HDLC Filter Device
            OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
            2,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
            OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
            OTX_FALSE,                 // If OTX_TRUE then all bits are inverted when transmitted and before received
            OTX_TRUE,                  // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
            OTX_HDLC_TR_NA,            // HDLC Transmission Rate for V.110 and X.30 modes
            OTX_TRUE                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
            ); DemoCheckRc(g_sqDspHdlcReceiver[i*4 + j].hHandle, result);          
        }
      }
      break;
    case OTX_DSP_SPM_OTXDTTX:
      // Nothing
      break;    
    default:
      DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }

    
    

    switch (g_demoEnv.eProgramModule) {
    case OTX_DSP_SPM_OTXSPM1:
    case OTX_DSP_SPM_OTXSPM2:
      // Initialize the DTMF Detectors (hosted in the DSP)
      result = OtxToneDtmfDetectorInit(g_sqDtmfDetector[i].hHandle, g_demoEnv.law, OTX_TRUE); DemoCheckRc(g_sqDtmfDetector[i].hHandle, result);

      // Initialize the Silence Detectors (hosted in the DSP)
      result = OtxToneSilenceDetectorInit(g_sqSilenceDetector[i].hHandle, g_demoEnv.law, 2, 2000); DemoCheckRc(g_sqSilenceDetector[i].hHandle, result);

      // Initialize the DTMF Dialers (hosted in the DSP)
      //result = OtxToneDtmfDiallerInit(g_sqDtmfDialer[i].hHandle, g_demoEnv.law, 12, 12, OTX_TONE_GAIN_DG_MINUS_6DB); DemoCheckRc(g_sqDtmfDialer[i].hHandle, result);
      //result = OtxToneDtmfDiallerInit(g_sqDtmfDialer[i].hHandle, g_demoEnv.law, 8, 8, OTX_TONE_GAIN_DG_MINUS_6DB); DemoCheckRc(g_sqDtmfDialer[i].hHandle, result);
      result = OtxToneDtmfDiallerInit(g_sqDtmfDialer[i].hHandle, g_demoEnv.law, 5, 5, OTX_TONE_GAIN_DG_MINUS_6DB); DemoCheckRc(g_sqDtmfDialer[i].hHandle, result);

      // Initialize the Data Converters (used to make a cross-connect within the DSP)
      result = OtxDataConverterInit(g_sqDataConverter[i].hHandle, g_demoEnv.law, g_demoEnv.law); DemoCheckRc(g_sqDataConverter[i].hHandle, result);

      break;
    case OTX_DSP_SPM_OTXDTTX:
      {
        // Initialize the DTMF Dialers (hosted in the DSP)
        OTX_INT32 j;
        for (j = 0; j < 32; j++) {
          if ((i*32 + j)%16 == 0) continue; // Skip 0 (sync), 16 (signalling), 32 (sync), 48 (signalling)
#if 0 // Higher gain. Longer tones
          result = OtxToneDtmfDiallerInit(g_sqDtmfDialer[i*32 + j].hHandle, g_demoEnv.law, 16, 16, OTX_TONE_GAIN_0DB); DemoCheckRc(g_sqDtmfDialer[i*32 + j].hHandle, result);
#else // Normal 
          result = OtxToneDtmfDiallerInit(g_sqDtmfDialer[i*32 + j].hHandle, g_demoEnv.law, 8, 8, OTX_TONE_GAIN_DG_MINUS_6DB); DemoCheckRc(g_sqDtmfDialer[i*32 + j].hHandle, result);
#endif
          result = OtxToneConstValGeneratorInit(g_sqConstValGenerator[i*32 + j].hHandle, OTX_DATA_FORMAT_UINT8, 0x55); DemoCheckRc(g_sqConstValGenerator[i*32 + j].hHandle, result);
          result = OtxToneValSamplerInit(g_sqValSampler[i*32 + j].hHandle, OTX_DATA_FORMAT_UINT8); DemoCheckRc(g_sqValSampler[i*32 + j].hHandle, result);
        }
      }
      break;
    case OTX_DSP_SPM_OTXHDLC:
      // Nothing
      break;    
    default:
      DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }
    
  }

}


void DemoEnableLogicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < 2; i++) {

    DemoLogf("Enable hHdlcSender[%d]\n", i);
    result = OtxDrvEnable(g_sqLiHdlcSender[i].hHandle,   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqLiHdlcSender[i].hHandle, result);

    DemoLogf("Enable hHdlcReceiver[%d]\n", i);
    result = OtxDrvEnable(g_sqLiHdlcReceiver[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqLiHdlcReceiver[i].hHandle, result);


    switch (g_demoEnv.eProgramModule) {
    case OTX_DSP_SPM_OTXSPM1:
    case OTX_DSP_SPM_OTXSPM2:
      DemoLogf("Enable hDspHdlcSender[%d]\n", i);
      result = OtxDrvEnable(g_sqDspHdlcSender[i].hHandle,   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
      
      DemoLogf("Enable hDspHdlcReceiver[%d]\n", i);
      result = OtxDrvEnable(g_sqDspHdlcReceiver[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDspHdlcReceiver[i].hHandle, result);
      
      DemoLogf("Enable hDtmfDialer[%d]\n", i);
      result = OtxDrvEnable(g_sqDtmfDialer[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDialer[i].hHandle, result);
      
      DemoLogf("Enable hDtmfDetector[%d]\n", i);
      result = OtxDrvEnable(g_sqDtmfDetector[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDetector[i].hHandle, result);

      DemoLogf("Enable hSilenceDetector[%d]\n", i);
      result = OtxDrvEnable(g_sqSilenceDetector[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqSilenceDetector[i].hHandle, result);

      DemoLogf("Enable hDataConverter[%d]\n", i);
      result = OtxDrvEnable(g_sqDataConverter[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDataConverter[i].hHandle, result);
      break;
    case OTX_DSP_SPM_OTXDTTX:
      {
        // Initialize the DTMF Dialers (hosted in the DSP)
        OTX_INT32 j;
        for (j = 0; j < 32; j++) {
          if (j%16 == 0) continue; // Skip 0 (sync), 16 (signalling), 32 (sync), 48 (signalling)
          //DemoLogf("Enable hDtmfDialer[%d]\n", i*32 + j);
          //result = OtxDrvEnable(g_sqDtmfDialer[i*32 + j].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDialer[i*32 + j].hHandle, result);
          DemoLogf("Enable hValSampler[%d]\n", i*32 + j);
          result = OtxDrvEnable(g_sqValSampler[i*32 + j].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDialer[i*32 + j].hHandle, result);
        }
      }
      break;
    case OTX_DSP_SPM_OTXHDLC:
      {
        OTX_INT32 j;
        for (j = 0; j < 4; j++) {
          DemoLogf("Enable hDspHdlcSender[%d]\n", i*4 + j);
          result = OtxDrvEnable(g_sqDspHdlcSender[i*4 + j].hHandle,   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDspHdlcSender[i*4 + j].hHandle, result);
          
          DemoLogf("Enable hDspHdlcReceiver[%d]\n", i*4 + j);
          result = OtxDrvEnable(g_sqDspHdlcReceiver[i*4 + j].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDspHdlcReceiver[i*4 + j].hHandle, result);
        }
      }
      break;    
    default:
      DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }
  }

}


/*******************************************************************************
*******************************************************************************/
static void DemoCrossConnect(
  OTX_INT32 nTsA,
  OTX_INT32 nTsB
)
{
  OTX_INT32 i;
  //OTX_INT32 j;
  OTX_INT32 ts;
  OTX_INT32 nLi = 0;
  OTX_INT32 nDsp = 0;
  OTX_INT32 nCapacity = 64;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_HANDLE hLiDspPipe[2];
  OTX_HANDLE hDspLiPipe[2];
  OTX_HANDLE hDataConverter[2]; // 2 for a duplex cross-connection
  
  assert((g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1) || (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2) || (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXHDLC));

#if 0
  {
    i = 0;
    DemoLogf("Connect DSP DTMF Detector\n");
    ts = 20;
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[ts] = 0xFF;
    nCapacity = 64;
    
    result = OtxDrvCreatePipe(nCapacity, &g_sqDtmfDetector[i].hPipe); DemoCheckRc(g_sqDtmfDetector[i].hPipe, result);
    result = OtxDrvConnectPipe(g_sqDtmfDetector[i].hPipe, OTX_PIPE_OUTPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDtmfDetector[i].hPipe, result);
    result = OtxDrvConnectPipe(g_sqDtmfDetector[i].hPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDtmfDetector[i].hPipe, result);
    
    result = OtxDrvConnectLogicalDevice(g_sqDtmfDetector[i].hPipe,  -1,  g_sqDtmfDetector[i].hHandle,    -1); DemoCheckRc(g_sqDtmfDetector[i].hPipe, result);
  }
#endif




  // Create pipes 
  for (i = 0; i < 2; i++) {
    result = OtxDrvCreatePipe(nCapacity, &hLiDspPipe[i]); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreatePipe(nCapacity, &hDspLiPipe[i]); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  }


  // Create one Data Converter device which will cross-connect 2 pipes (duplex)
  for (i = 0; i < 2; i++) {
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_DATA_CONVERTER, g_hEventQueue, 1000+i, &hDataConverter[i]); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  }
  
    
  DemoLogf("Cross-Connect TS%d <-> TS%d\n", nTsA, nTsB);
  
  // Connect 2 pipes between the Li and the DSP
  for (i = 0; i < 2; i++) {
    if (i == 0) {
      ts = nTsA;
    } else {
      ts = nTsB;
    }
    
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[ts] = 0xFF;
    
    result = OtxDrvConnectPipe(hDspLiPipe[i], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvConnectPipe(hDspLiPipe[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],   -1, sqTimeSlotMask, 32); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    result = OtxDrvConnectPipe(hLiDspPipe[i], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],   -1, sqTimeSlotMask, 32); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvConnectPipe(hLiDspPipe[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  }
  
  // nTsA -> nTsB  (hDataConverter[0])
  result = OtxDrvConnectLogicalDevice(hLiDspPipe[0],      -1, hDataConverter[0], -1); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  result = OtxDrvConnectLogicalDevice(hDataConverter[0],  -1, hDspLiPipe[1],     -1); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  
  // nTsB -> nTsA  (hDataConverter[1])
  result = OtxDrvConnectLogicalDevice(hLiDspPipe[1],      -1, hDataConverter[1], -1); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  result = OtxDrvConnectLogicalDevice(hDataConverter[1],  -1, hDspLiPipe[0],     -1); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  
  result = OtxDataConverterInit(hDataConverter[0], OTX_DATA_FORMAT_UINT8, OTX_DATA_FORMAT_UINT8); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  result = OtxDataConverterInit(hDataConverter[1], OTX_DATA_FORMAT_UINT8, OTX_DATA_FORMAT_UINT8); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  
  result = OtxDrvEnable(hDataConverter[0], OTX_TASK_SYNC, 5000L); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  result = OtxDrvEnable(hDataConverter[1], OTX_TASK_SYNC, 5000L); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  
  DemoLogf("Connection completed\n");
  
  OsaThreadSleep(2000);
  
#if 1
  DemoLogf("Break Cross-Connect TS%d <-> TS%d\n", nTsA, nTsB);
  
  result = OtxDrvDisable(hDataConverter[0], OTX_TASK_SYNC, 5000L); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  result = OtxDrvDisable(hDataConverter[1], OTX_TASK_SYNC, 5000L); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  
  // TS11 -> TS12  (hDataConverter[0])
  result = OtxDrvDisconnectLogicalDevice(hLiDspPipe[0],      -1, hDataConverter[0], -1); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  result = OtxDrvDisconnectLogicalDevice(hDataConverter[0],  -1, hDspLiPipe[1],     -1); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  
  // TS12 -> TS11  (hDataConverter[1])
  result = OtxDrvDisconnectLogicalDevice(hLiDspPipe[1],      -1, hDataConverter[1], -1); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  result = OtxDrvDisconnectLogicalDevice(hDataConverter[1],  -1, hDspLiPipe[0],     -1); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  
  // Connect 2 pipes on TS11 and TS12 between the Li and the DSP
  for (i = 0; i < 2; i++) {
    result = OtxDrvDisconnectPipe(hDspLiPipe[i], OTX_PIPE_INPUT  ); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvDisconnectPipe(hDspLiPipe[i], OTX_PIPE_OUTPUT ); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    result = OtxDrvDisconnectPipe(hLiDspPipe[i], OTX_PIPE_INPUT  ); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvDisconnectPipe(hLiDspPipe[i], OTX_PIPE_OUTPUT ); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  }
  
  DemoLogf("Connection broken\n");
#endif

  // Cleanup
  for (i = 0; i < 2; i++) {
    result = OtxDrvClose(hDataConverter[i]); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvClose(hLiDspPipe[i]); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvClose(hDspLiPipe[i]); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  }

  OsaThreadSleep(2000);

}


void DemoInitT1E1ConfigData(
/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 128;
  pE1CfgData->nLosRecovery               = 128;
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF; // OTX_T1_FF_F12
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 128;
  pT1CfgData->nLosRecovery               = 128;
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_FALSE;
  pT1CfgData->bfClearChannels            = 0x00;
}




/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
#ifdef DEMO_USE_TIMESTAMP_FROM_1601 // Timestamps from Jan 1, 1601

  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampUs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  // Calculate number of microseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampUs.nUint64 = nTimeStamp.nUint64 / 10 - (nTimeStampSec.nUint64 * 1000000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.6d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampUs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }

#else // Timestamp using QueryPerformanceCounter(), which start counting from PC boot-time
  OTX_DATETIME nTimeStampTemp;
  OTX_UINT32 nSecs;
  OTX_UINT32 nMilliSecs;
  OTX_UINT32 nMicroSecs;
  OTX_UINT32 nNanoSecs;
  
  nTimeStampTemp.nUint64 = nTimeStamp.nUint64;
  
  nSecs = (OTX_UINT32) (nTimeStampTemp.nUint64 / g_nFrequency);
  nTimeStampTemp.nUint64 %= g_nFrequency;
  nTimeStampTemp.nUint64 *= 1000;
  nMilliSecs = (OTX_UINT32) (nTimeStampTemp.nUint64 / g_nFrequency);
  nTimeStampTemp.nUint64 %= g_nFrequency;
  nTimeStampTemp.nUint64 *= 1000;
  nMicroSecs = (OTX_UINT32) (nTimeStampTemp.nUint64 / g_nFrequency);
  nTimeStampTemp.nUint64 %= g_nFrequency;
  nTimeStampTemp.nUint64 *= 1000;
  nNanoSecs = (OTX_UINT32) (nTimeStampTemp.nUint64 / g_nFrequency);
  
  //DemoLogf("%u.%.3u,%.3u,%.3u (%d*2^32 + %d)", nSecs, nMilliSecs, nMicroSecs, nNanoSecs, nTimeStamp.sUint32.nHigh, nTimeStamp.sUint32.nLow);
  DemoLogf("%u.%.3u,%.3u,%.3u ", nSecs, nMilliSecs, nMicroSecs, nNanoSecs);

#endif
  
}

/*******************************************************************************
  @func
    Raw Data Sender Thread Function 
*******************************************************************************/
void *DemoDtmfDialer(
  void *lpParameter   // thread data
)
{
  OTX_DATETIME nTimeStamp;
  OtxEventDataS eventData;
  OTX_RESULT result;
  OTX_RESULT otxRc;
  char msgStr[256];
#if !defined(_WIN32_WCE)
  clock_t startTime, stopTime;
#endif
  OTX_UINT32 nDialer = *((OTX_INT32 *) lpParameter);
  OTX_INT32 nDigits;
  OTX_INT32 i;
  OTX_BOOL bDialerComplete;


#if !defined(_WIN32_WCE)
  startTime = clock();
#endif

  //DemoLogf("nDialer = %d\n", nDialer);
  if (nDialer >= 64) {
    DemoLogf("Error: DemoDtmfDialer Thread was passed an invalid nDialer = %d. Thread ending...\n", nDialer);
    goto END;
  }


  for (;;) { // loop forever

    while (!g_sqDtmfDialer[nDialer].bActive) {
      OsaThreadSleep(500);
    }

    g_sqDtmfDialer[nDialer].coDtmfCompleted = 0;
    
    for (i = 0; i < g_sqDtmfDialer[nDialer].nRepeatTimes; i++) {
      
      // Discard any events from  event queue (if any)
      do {
        result = OtxDrvGetEventData(g_sqDtmfDialer[nDialer].hEventQueue, &eventData); 
      } while (result == OTX_S_OK);
      
      
      if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXDTTX) {
        result = OtxDrvEnable(g_sqDtmfDialer[nDialer].hHandle,  OTX_TASK_SYNC, 5000L); 
        
        if (result != OTX_S_OK) {
          
          if (OTX_RESULT_CODE(result) == OTXDSP_SDK_E_BUSY) {
            DemoLogf("DTMF Dialer could not be enabled because another device is occupying this channel. Disabling ConstValGenerator\n");
            result = OtxDrvDisable(g_sqConstValGenerator[nDialer].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqConstValGenerator[nDialer].hHandle, result);
            if (result != OTX_S_OK) {
              DemoLogf("Could not disable the ConstValGenerator. Aborting this operation\n");
              goto END;
            } 
            DemoLogf("Attempting to enable the DTMF Dialer (now when the ConstValGenerator is disabled\n");
            result = OtxDrvEnable(g_sqDtmfDialer[nDialer].hHandle,  OTX_TASK_SYNC, 5000L); 
            if (result != OTX_S_OK) {
              DemoLogf("Could not enable the DTMF dialler. Aborting this operation\n");
              goto END;
            } 
            
          } else {
            DemoLogf("Could not enable the DTMF dialler (cause unknown)\n");
            DemoCheckRc(g_sqDtmfDialer[nDialer].hHandle, result);
            goto END;
          }
        }
      }
      
      
      nDigits = strlen(g_sqDtmfDialer[nDialer].sqDigits);
      if (g_sqDtmfDialer[nDialer].nRepeatTimes == 1) {
        OtxDrvGetCurrentTime(&nTimeStamp);
        DemoPrintTimeStamp(nTimeStamp);
        DemoLogf(" DtmfDialer[%d] Dialling digits %s.\n", nDialer, g_sqDtmfDialer[nDialer].sqDigits);
      }
      
      // Start Dialing DTMF number, Non-Blocking call, will receive Event when completed
      result = OtxToneDtmfDial(g_sqDtmfDialer[nDialer].hHandle, (OTX_UINT8 *)g_sqDtmfDialer[nDialer].sqDigits, nDigits, DEMO_TASK_REF, OTX_TIME_INFINITY); 
      if (OTX_RESULT_CODE(result) == OTX_DTMF_DIALER_RES_BUSY) {
        DemoLogf("Dialer busy. Aborting dialler thread\n");
        goto END;
      } else if (result != OTX_S_OK) {
        DemoCheckRc(g_sqDtmfDialer[nDialer].hHandle, result);
      }
      
      
      // Wait for events from the Dtmf Dialer
      bDialerComplete = OTX_FALSE;
      while (!bDialerComplete) {
        OtxDrvWaitForSingleEvent(g_sqDtmfDialer[nDialer].hNotificationEvent, OTX_TIME_INFINITY);
        do {
          result = OtxDrvGetEventData(g_sqDtmfDialer[nDialer].hEventQueue, &eventData); DemoCheckRc(g_sqDtmfDialer[nDialer].hEventQueue, result);
          
          if (result == OTX_S_OK) {
            // We have a driver event
            switch (eventData.m_eDeviceType) {
              
            case OTX_LDEVICE_TONE_DTMF_DIALER:
              otxRc = OtxDrvEventCode2String(eventData.m_hDevice, eventData.m_nCode, 256, &msgStr[0]);
              if (OTX_SUCCESS(otxRc)) {
                if (g_sqDtmfDialer[nDialer].nRepeatTimes == 1) {
                  DemoLogf("DtmfDialer[%d] Event: %s \n", eventData.m_nSourceId, msgStr);
                }
              } else {
                DemoLogf("DtmfDialer[%d] Invalid Event: %ld \n", eventData.m_nSourceId, eventData.m_nCode);
              }
              
              if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXDTTX) {
                if (eventData.m_nCode == OTX_DTMF_DIALER_EC_COMPLETE) {
                  if (g_sqDtmfDialer[nDialer].nRepeatTimes == 1) {
                    OtxDrvGetCurrentTime(&nTimeStamp);
                    DemoPrintTimeStamp(nTimeStamp);
                    DemoLogf(" DtmfDialer[%d] Disabling and reseting device\n", eventData.m_nSourceId);
                  }
                  g_sqDtmfDialer[nDialer].coDtmfCompleted++;
                  
                  result = OtxDrvDisable(g_sqDtmfDialer[eventData.m_nSourceId].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDialer[eventData.m_nSourceId].hHandle, result);
                  result = OtxDrvReset(g_sqDtmfDialer[eventData.m_nSourceId].hHandle); DemoCheckRc(g_sqDtmfDialer[eventData.m_nSourceId].hHandle, result);
                  bDialerComplete = OTX_TRUE;
                }
              }
              break;
              
            default:
              // Unhandled event
              DemoLogf("DemoDtmfDialer[%d]: Event from an unknown object 0x%X\n", nDialer, eventData.m_hDevice);
              break;
            }
            
          }

          if (g_sqDtmfDialer[nDialer].bAbort) {
            DemoLogf("DemoDtmfDialer[%d]: Thread aborted\n", nDialer);
            goto END;
          }

        } while (result == OTX_S_OK);
        
      } // !bDialerComplete

    }
    
END:
    
#if !defined(_WIN32_WCE)
    stopTime = clock();
    DemoLogf("DemoDtmfDialer[%d]: Done (time=%2.1f seconds, coDtmfCompleted=%d)\n", nDialer, (double)(stopTime - startTime) / CLOCKS_PER_SEC, g_sqDtmfDialer[nDialer].coDtmfCompleted);
#endif

    g_sqDtmfDialer[nDialer].bActive = OTX_FALSE;
        
  } // loop forever

  return 0;

}



/*******************************************************************************
  @func
    Raw Data Sender Thread Function 
*******************************************************************************/
void *DemoAutoTransmitHdlcFunction(
  void *lpParameter   // thread data
)
{
  OTX_INT32 i;
  OTX_INT32 busyCount = 0;
#if !defined(_WIN32_WCE)
  clock_t startTime, stopTime;
#endif
  OTX_RESULT result;

  OTX_INT32 nCurrentLi = *((OTX_INT32 *) lpParameter);

  assert(nCurrentLi <= 2);
  
#if !defined(_WIN32_WCE)
  startTime = clock();
#endif
    
  while (g_bAutoTransmit[nCurrentLi]) {
    
    if (g_cbLiHdlcSenderFrame > OD_HDLC_MAX_TX_HDLC_FRAME_LEN) {
      DemoLogf("Error: Frame Length too long\n");
      break;
    }
    
    g_cbHdlcFrame = g_cbLiHdlcSenderFrame;
    for (i = 0; i < g_cbHdlcFrame; i++) {
      g_sqHdlcFrame[i] = (OTX_UINT8) i;
    }
    
    
    // Send HDLC Frame(s)
    for (i = 0; i < g_coHdlcSenderFrames; i++) {
      result = OtxHdlcSendData(g_sqLiHdlcSender[nCurrentLi].hHandle, g_sqHdlcFrame, g_cbHdlcFrame, DEMO_TASK_REF, 200); 
      if (result != OTX_S_OK) {
        DemoLogf("DemoAutoTransmitHdlcFunction[%d] OtxHdlcSendData() returned rc=%X\n", nCurrentLi, result);
        goto ENDIT;
      }
    }

    OsaThreadSleep(950);
    
  }

ENDIT:

#if !defined(_WIN32_WCE)
  stopTime = clock();
  DemoLogf("DemoAutoTransmitHdlcFunction[%d] thread terminated (BusyCount=%d, time=%2.1f seconds).\n", nCurrentLi, busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
#endif

  return 0;
}




/*******************************************************************************
  @func

*******************************************************************************/
void  *DemoTimeStampTest(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_DATETIME nTimeStamp;
  
  for (;;) {
    
    result = OtxDrvGetCurrentTime(&nTimeStamp); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

    DemoPrintTimeStamp(nTimeStamp);
    DemoLogf("\n");
    
    OsaThreadSleep(2000);
  }
  


  return 0;

}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (result == OTX_S_OK) {
        // We have a driver event
        // Print the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (result == OTX_S_OK);
  }

  return 0;

}


void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  //OTX_RESULT result;
  OTX_UINT32 i;
  OTX_UINT8  buf[2048];
  OTX_UINT32 nFrameLen;
  OTX_UINT32 nFrameSeq;
  OtxHdlcFrameEndCodeE nFrameEndCode;
  OTX_DATETIME nTimeStamp;
  char msgStr[256];

  //OsaMutexLock(&g_analyzeCritSec);

  //OtxDrvGetCurrentTime(&nTimeStamp);
  //DemoLogf("%u%u We have an event:\n", nTimeStamp.sUint32.nHigh, nTimeStamp.sUint32.nLow/10000);
  
  switch (pEventData->m_eDeviceType) {
  case OTX_LDEVICE_HDLC_RECEIVER:
    if (g_bHdlcVerbose) {
      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    }
    break;
  default:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    break;
  }

    
  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_THOR_8_PCI:
    // Event from the board
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("Brd[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("Brd[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("Brd[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }

    /*
    if (pEventData->m_nCode == OTX_BRD_EC_AUTO_DISABLED) {
      OTX_INT32 nLi;
      OTX_INT32 nDsp;
      // The board device encountered a severe error (ESD pulse?). Try to revive it by Reseting the Falc devices and Enabling them
      OsaThreadSleep(100); // Wait for specified time to before attempting to reset the device

      for (nLi = 0; nLi < 2; nLi++) {
        DemoLogf("Reseting Li%d\n", nLi);
        otxRc = OtxDrvReset(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], otxRc);
      }

      for (nLi = 0; nLi < 2; nLi++) {
        DemoLogf("Enabling Li%d\n", nLi);
        otxRc = OtxDrvEnable(g_hPhysDevs.hLi[nLi], DEMO_TASK_REF, 500); DemoCheckRc(g_hPhysDevs.hLi[nLi], otxRc);
      }

      for (nDsp = 0; nDsp < 2; nDsp++) {
        DemoLogf("Enabling Dsp%d\n", nDsp);
        otxRc = OtxDrvEnable(g_hPhysDevs.hDsp[nDsp], DEMO_TASK_REF, 500); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], otxRc);
      }
    }
    */
    break;

  case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    
    /*
    if (pEventData->m_nCode == OTX_T1E1_EC_AUTO_DISABLED) {
      // The T1E1 device encountered a severe error (ESD pulse?). Try to revive it by Reseting and Enabling it.
      OsaThreadSleep(100); // Wait for specified time to before attempting to reset the device
      DemoLogf("E1/T1[%d] Reseting device \n", pEventData->m_nSourceId);
      otxRc = OtxDrvReset(pEventData->m_hDevice); DemoCheckRc(pEventData->m_hDevice, otxRc);
      DemoLogf("E1/T1[%d] Enabling device \n", pEventData->m_nSourceId);
      otxRc = OtxDrvEnable(pEventData->m_hDevice, DEMO_TASK_REF, 500); DemoCheckRc(pEventData->m_hDevice, otxRc);
    }
    */
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;
    
  case OTX_LDEVICE_TONE_DTMF_DIALER:
    if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXDTTX) {
      DemoLogf("DtmfDialer[%d] Unexpected Event (to the wrong thread)\n", pEventData->m_nSourceId);
    } else {
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("DtmfDialer[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("DtmfDialer[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
      
    }

    break;

  case OTX_LDEVICE_TONE_DTMF_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dtmf Detector 
      if (pEventData->m_nCode == OTX_DTMF_DETECTOR_EC_DIGIT) {
        switch (pEventData->m_nParam) {
        case 0:  // 0  (941Hz/1336Hz)
        case 1:  // 1  (697Hz/1209Hz)
        case 2:  // 2  (697Hz/1336Hz)
        case 3:  // 3  (697Hz/1477Hz)
        case 4:  // 4  (770Hz/1209Hz)
        case 5:  // 5  (770Hz/1336Hz)
        case 6:  // 6  (770Hz/1477Hz)
        case 7:  // 7  (852Hz/1209Hz)
        case 8:  // 8  (852Hz/1336Hz)
        case 9:  // 9  (852Hz/1477Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF digit %d \n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        case 10: // A  (697Hz/1633Hz)
        case 11: // B  (770Hz/1633Hz)
        case 12: // C  (852Hz/1633Hz)                        ``
        case 13: // D  (941Hz/1633Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF key %C \n", pEventData->m_nSourceId, pEventData->m_nParam - 10 + 'A');
          break;
        case 14: // *  (941Hz/1209Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF key '*'\n", pEventData->m_nSourceId);
          break;
        case 15: // #  (941Hz/1477Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF key '#'\n", pEventData->m_nSourceId);
          break;
        default: 
          DemoLogf("DtmfDetector[%d] Unrecognized DTMF tone (%d)\n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        }

      } else {
        // Unexpected event from this device
        DemoLogf("DtmfDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("DtmfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_HDLC_RECEIVER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from HLDC Receiver 
      if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
        
        otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
        while (otxRc == OTX_S_OK) {

          if (pEventData->m_nSourceId < 256) {
            g_coHdlcRxFrames[pEventData->m_nSourceId]++;
          }

          if (g_bHdlcVerbose) {
            DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);

            DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
            for (i = 0; i < nFrameLen && i < 1024; i++) {
              DemoLogf("%.2X ", buf[i]);
            }
            DemoLogf("\n");
          } else {
            if (nFrameEndCode) {

              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("FrameEndCode Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
              for (i = 0; i < nFrameLen && i < 256; i++) {
                DemoLogf("%.2X ", buf[i]);
              }
              DemoLogf("\n");
              
              if (g_coHdlcBadRxFrames[pEventData->m_nSourceId] > 100) {
                DemoLogf("Too many bad frames from HdlcReceiver[%d]. Disabling device\n", pEventData->m_nSourceId);
                otxRc = OtxDrvDisable(pEventData->m_hDevice,  OTX_TASK_SYNC, 5000L); DemoCheckRc(pEventData->m_hDevice, otxRc);
              }


            } else {
              if ((nFrameLen != (OTX_UINT32)g_cbHdlcFrame) && (nFrameLen != 3)) {
                g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
                
                DemoLogf("LENGTH Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
                for (i = 0; i < nFrameLen && i < 256; i++) {
                  DemoLogf("%.2X ", buf[i]);
                }
                DemoLogf("\n");
                
              } else if (memcmp(buf+3, g_sqHdlcFrame+3, g_cbHdlcFrame-3)) {
                g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
                
                DemoLogf("memcmp Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
                for (i = 0; i < nFrameLen && i < 256; i++) {
                  DemoLogf("%.2X ", buf[i]);
                }
                DemoLogf("\n");
                
              }
            }

          }
          otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
        }
        if (otxRc != OTX_S_FALSE) {
          DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful (rc = %X).\n", pEventData->m_nSourceId, otxRc);
        }

      } else {
        // Unexpected event from this device
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;
    
  case OTX_LDEVICE_HDLC_SENDER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


  //OsaMutexUnlock(&g_analyzeCritSec);

}






/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->law               = OTX_DATA_FORMAT_COMPAND8_A_LAW;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->eT1E1Mode         = OTX_T1E1_MODE_T1;
  envP->eClkSource        = OTX_CLOCK_SOURCE_LOCAL_0;
  envP->eProgramModule    = OTX_DSP_SPM_OTXHDLC;
  envP->nBoardType        = OTX_DEVICE_THOR_PCMCIA;
  envP->bUseDspDebugger   = OTX_FALSE;
  envP->coDsp             = 2;


  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'D') {   // DSP Debugger
          envP->bUseDspDebugger = OTX_TRUE;
          envP->coDsp             = 4;
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '2') {
            envP->nBoardType = OTX_DEVICE_THOR_PCI;
            envP->coLi = 2;
          } else if ( *(argv[i]+2) == 'M') {
            envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
            envP->coLi = 2;
          } else if ( *(argv[i]+2) == '8') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI;
            envP->coLi = 8;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        
        } else if (*(argv[i]+1) == 'P') {     // Dsp Program module
          if (*(argv[i]+2) == '1') {          // OtxSpm1.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM1;
          } else if (*(argv[i]+2) == '2') {   // OtxSpm2.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2;
          } else if (*(argv[i]+2) == 'D') {   // OtxDtTx.out
            envP->eProgramModule = OTX_DSP_SPM_OTXDTTX;
          } else if (*(argv[i]+2) == 'H') {   // OtxHdlc.out
            envP->eProgramModule = OTX_DSP_SPM_OTXHDLC;
          } else {
            DemoLogf("Error: Invalid DSP Program Module command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'C') {   // Companding Law
          if (*(argv[i]+2) == 'A') {   // A-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'U') {   // U-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'K') {   // Clock Source
          if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == '0') {   // Line Interface 0
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
          } else if (*(argv[i]+2) == '1') {   // Line Interface 1
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
          } else if (*(argv[i]+2) == 'E') {   // External
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_2;
          } else if (*(argv[i]+2) == '3') {   // Split clock, both Lis synchronized to its own line
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_3;
          } else if (*(argv[i]+2) == '4') {   // Split clock, Li0 synchronized to its T1E1 span. Li1 internal
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_4;
          } else if (*(argv[i]+2) == '5') {   // Split clock, Li1 synchronized to its T1E1 span. Li0 internal
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_5;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        exit(1);
      }
    }
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[2|8|M]    : Board Type used for the demo, where\n");
  DemoLogf("                  2 = Use Thor-2-PCI Board with Vidar-5x4-ASM\n");
  DemoLogf("                  8 = Use Thor-8-PCI Board\n");
  DemoLogf("                  M = Use Thor-2-PCMCIA Board\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /KI          : Use Internal board oscillator as clock source\n");
  DemoLogf(" /K0          : Use recovered clock from Line Interface 0 as clock source\n");
  DemoLogf(" /K1          : Use recovered clock from Line Interface 1 as clock source\n");
  DemoLogf(" /K3          : Split Clock, both Line Interface are synchronized to its T1E1 span\n");
  DemoLogf(" /K4          : Split Clock, Li0 synchronized to its T1E1 span. Li1 internal\n");
  DemoLogf(" /K5          : Split Clock, Li1 synchronized to its T1E1 span. Li0 internal\n");
  DemoLogf(" /KE          : Use External clock source\n");
  DemoLogf(" /CA          : A-law\n");
  DemoLogf(" /CU          : u-law\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /P1          : Use OtxSpm1 DSP module (instead of OtxHdlc as default)\n");
  DemoLogf(" /P2          : Use OtxSpm2 DSP module (instead of OtxHdlc as default)\n");
  DemoLogf(" /PD          : Use OtxDtTx DSP module (instead of OtxHdlc as default)\n");
  DemoLogf(" /PH          : Use OtxHdlc DSP module (default)\n");
  DemoLogf(" /D           : DSP Debugger mode. Prompt to start DSP code from Debugger\n");
  DemoLogf(" /Z<str>      : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);
}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    OsaBeep(4000, 200);
    DemoLogf("\n");
    OtxDrvDbgDump(g_hEventQueue, 0);

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d): %s \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      } else {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d) Cannot translate this result code \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber);
      }
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
{
  static OTX_INT32 nCurrentLi = 0;
  static OTX_INT32 nCurrentHdlcSender = 0;
  static OTX_INT32 nCurrentDialer = 1;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-1 Select active Li or Dsp, this Li/Dsp is then used for subsequent commands\n");
    //DemoLogf(" o   Debug command to the driver\n");
    DemoLogf(" t   Send a HDLC frame on TS16 (E1) or TS24 (T1) (using HDLC Sender in the Li)\n");

    switch (g_demoEnv.eProgramModule) {
    case OTX_DSP_SPM_OTXSPM1:
    case OTX_DSP_SPM_OTXSPM2:
    case OTX_DSP_SPM_OTXHDLC:
      DemoLogf(" T   Send a HDLC frame on TS1  (using HDLC Sender in the DSP)\n");
      DemoLogf(" N   Set HDLC Frame length for DSP HDLC Sender\n");
      break;
    case OTX_DSP_SPM_OTXDTTX:
      // Nothing
      break;
    default:
      DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }

    switch (g_demoEnv.eProgramModule) {
    case OTX_DSP_SPM_OTXSPM1:
    case OTX_DSP_SPM_OTXSPM2:
      DemoLogf(" d   Dial DTMF digits on TS 20 from the current Li\n");
      DemoLogf(" D   Dial DTMF digits on TS 20 from the current Li (and check for silence)\n");
      DemoLogf(" |   Modify DTMF Detector minimum duration\n");
      //DemoLogf(" &   Use the Data Converter to make cross-connects on the current Dsp/Li (%d)\n", nCurrentLi);
      break;
    case OTX_DSP_SPM_OTXDTTX:
      DemoLogf(" d   Dial DTMF digits from the current Dialer\n");
      DemoLogf(" m   Increment Dialer (currently Dialer#%d (Li%d, TS%d) Selected\n", nCurrentDialer, nCurrentDialer/32, nCurrentDialer%32);
      DemoLogf(" M   Decrement Dialer (currently Dialer#%d (Li%d, TS%d) Selected\n", nCurrentDialer, nCurrentDialer/32, nCurrentDialer%32);
      DemoLogf(" D   Dial DTMF digits on all DTMF Dialers\n");
      DemoLogf(" )   Turn on  Constant value on all timeslots (constval=TS)\n");
      DemoLogf(" c   Turn on  Constant value on a selected timeslot\n");
      DemoLogf(" C   Turn off Constant value on a selected timeslot\n");
      DemoLogf(" s   Read sample value from a selected timeslot\n");
      DemoLogf(" S   Read sample value from a selected timeslot\n");
      break;
    case OTX_DSP_SPM_OTXHDLC:
      // Nothing
      DemoLogf(" m   Increment DSP HDLC Sender device (currently Dev#%d selected\n", nCurrentHdlcSender);
      DemoLogf(" M   Decrement DSP HDLC Sender device (currently Dev#%d selected\n", nCurrentHdlcSender);
      break;    
    default:
      DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }

    DemoLogf(" l   Query Line Status of both Lis\n");
    DemoLogf(" n   Set HDLC Frame length for Li HDLC Sender\n");
    DemoLogf(" +   Set number of consecutive frame to be send from the HDLC Sender\n");
    DemoLogf(" a   Start Sending AIS (All 1's) on current Li\n");
    DemoLogf(" A   Stop Sending AIS (All 1's) on current Li\n");
    DemoLogf(" e/E Enable/Disable current Dsp\n");
    DemoLogf(" y/Y Enable/Disable current Li\n");
    DemoLogf(" .   Toggle OTX_ATTR_T1E1_SECONDS_INTERRUPT attribute\n");
    DemoLogf(" k   Toggle OTX_ATTR_T1E1_AUTOMATIC_RESET attribute\n");
    DemoLogf(" K   Reset Board\n");
    DemoLogf(" g   Reset current Dsp\n");
    DemoLogf(" G   Reset current Li\n");
    DemoLogf(" ;   Toggle Automatic transmitting of HDLC frames on/off for current Li\n");
    break;

  case '|':
    {
      OTX_UINT32 nMinimumToneDuration;
      OTX_INT32 i;
      char strBuf[12];
      
      result = OtxDrvGetAttributeValueUint32(g_sqDtmfDetector[0].hHandle, OTX_ATTR_DTMF_DETECTOR_MIN_DURATION, &nMinimumToneDuration); DemoCheckRc(g_sqDtmfDetector[0].hHandle, result);
      DemoLogf("Please enter minimum tone duration for DTMF digits [%d milliseconds]", nMinimumToneDuration);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nMinimumToneDuration = (OTX_UINT32) strtol(strBuf, NULL, 0);
      }
      
      DemoLogf("Setting minimum tone duration = %d milliseconds for all DTMF detectors\n", nMinimumToneDuration);
      for (i = 0; i < 2; i++) {
        
        result = OtxDrvDisable(g_sqDtmfDetector[i].hHandle, OTX_TASK_SYNC, 500L); DemoCheckRc(g_sqDtmfDetector[i].hHandle, result);
        
        result = OtxDrvSetAttributeValueUint32(g_sqDtmfDetector[i].hHandle, OTX_ATTR_DTMF_DETECTOR_MIN_DURATION, nMinimumToneDuration); DemoCheckRc(g_sqDtmfDetector[i].hHandle, result);
        
        result = OtxToneDtmfDetectorInit(g_sqDtmfDetector[i].hHandle, g_demoEnv.law, OTX_TRUE); DemoCheckRc(g_sqDtmfDetector[i].hHandle, result);
        
        result = OtxDrvEnable(g_sqDtmfDetector[i].hHandle, OTX_TASK_SYNC, 500L); DemoCheckRc(g_sqDtmfDetector[i].hHandle, result);
        
      }
      
    }
    break;


  case ':':
    {
      OTX_INT32 i;

      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        DemoLogf("Switching both Lis to OTX_T1E1_MODE_E1 mode\n");
        g_demoEnv.eT1E1Mode = OTX_T1E1_MODE_E1;
      } else {
        DemoLogf("Switching both Lis to OTX_T1E1_MODE_T1 mode\n");
        g_demoEnv.eT1E1Mode = OTX_T1E1_MODE_T1;
      }


      for (i = 0; i < 2; i++) {
        
        // Set termination for the T1/E1 Front End
        result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        
        // Configure Transceivers
        if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
          // T1 mode
          result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
        } else {
          // E1 mode
          result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
        }
        result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        
      }

      DemoLogf("Done\n");
  
    }
    break;

  case '\'':
    {
      // Disable/Enable DSP Serial port clocks 
      OTX_BOOL bDspPortsDisabled;
      
      result = OtxDrvGetAttributeValueBool(
        g_hPhysDevs.hThorController0, 
        OTX_ATTR_BRD_DSP_PORTS_DISABLED, 
        &bDspPortsDisabled
        ); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      
      bDspPortsDisabled = !bDspPortsDisabled;
      DemoLogf("Setting OTX_ATTR_BRD_DSP_PORTS_DISABLED attribute to %s.\n", bDspPortsDisabled ? "TRUE":"FALSE");
      
      result = OtxDrvSetAttributeValueBool(
        g_hPhysDevs.hThorController0, 
        OTX_ATTR_BRD_DSP_PORTS_DISABLED, 
        bDspPortsDisabled
        ); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
    
  case '.':
    {
      OTX_BOOL bGenerateSecondsInterrupt;

      result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_SECONDS_INTERRUPT, &bGenerateSecondsInterrupt); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      bGenerateSecondsInterrupt = !bGenerateSecondsInterrupt;
      DemoLogf("Setting OTX_ATTR_T1E1_SECONDS_INTERRUPT attribute to %s for Li #%d\n", bGenerateSecondsInterrupt ? "TRUE":"FALSE", nCurrentLi);
      
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_SECONDS_INTERRUPT, bGenerateSecondsInterrupt); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    }
    break;

  case 'k':
    {
      OTX_BOOL bAutomaticResetMode;

      result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_AUTOMATIC_RESET, &bAutomaticResetMode); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      bAutomaticResetMode = !bAutomaticResetMode;
      DemoLogf("Setting OTX_ATTR_T1E1_AUTOMATIC_RESET attribute to %s for Li #%d\n", bAutomaticResetMode ? "TRUE":"FALSE", nCurrentLi);
      
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_AUTOMATIC_RESET, bAutomaticResetMode); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    }
    break;

  case 'K':
    {
      DemoLogf("Reseting Board \n");
      result = OtxDrvReset(g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      DemoLogf("Done\n");
    }
    break;

  case 'G':
    {
      DemoLogf("Reseting Li #%d \n", nCurrentLi);
      result = OtxDrvReset(g_hPhysDevs.hLi[nCurrentLi]); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      DemoLogf("Done\n");
    }
    break;

  case 'g':
    {
      DemoLogf("Reseting DSP #%d \n", nCurrentLi);
      result = OtxDrvReset(g_hPhysDevs.hDsp[nCurrentLi]); DemoCheckRc(g_hPhysDevs.hDsp[nCurrentLi], result);
      DemoLogf("Done\n");
    }
    break;

  case 'e':
    {
      DemoLogf("Enable Dsp #%d \n", nCurrentLi);
      result = OtxDrvEnable(g_hPhysDevs.hDsp[nCurrentLi], OTX_TASK_SYNC, 500L); DemoCheckRc(g_hPhysDevs.hDsp[nCurrentLi], result);
      DemoLogf("Done\n");
    }
    break;

  case 'E':
    {
      DemoLogf("Disable Dsp #%d \n", nCurrentLi);
      result = OtxDrvDisable(g_hPhysDevs.hDsp[nCurrentLi], OTX_TASK_SYNC, 500L); DemoCheckRc(g_hPhysDevs.hDsp[nCurrentLi], result);
      DemoLogf("Done\n");
    }
    break;


  case 'y':
    {
      DemoLogf("Enable Li #%d \n", nCurrentLi);
      result = OtxDrvEnable(g_hPhysDevs.hLi[nCurrentLi], OTX_TASK_SYNC, 500L); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      DemoLogf("Done\n");
    }
    break;

  case 'Y':
    {
      DemoLogf("Disable Li #%d \n", nCurrentLi);
      result = OtxDrvDisable(g_hPhysDevs.hLi[nCurrentLi], OTX_TASK_SYNC, 500L); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      DemoLogf("Done\n");
    }
    break;


  case '&':
    {
      OTX_INT32 i;
      static OTX_INT32 nIterations = 1;
      char strBuf[12];

      DemoLogf("How many iterations (%d)\n", nIterations);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nIterations = (OTX_INT32) strtol(strBuf, NULL, 0);
      }


      DemoLogf("Running DemoCrossConnect() (nIterations=%d)\n", nIterations);

      for (i = 0; i < nIterations; i++) {
        DemoLogf("Loop counter %d \n\n", i);
        DemoCrossConnect(11, 12);
        DemoCrossConnect(13, 14);
      }
    }
    break;


#if 1 // Only for debugging
  case 'o':
    {
      char szBuf[12];
      OTX_INT32 nCmd;
      
      DemoLogf("Driver Debug Command. Which Cmd?\n");
      DemoLogf("1: Disable DPC\n");
      DemoLogf("2: Enable DPC\n");
      DemoLogf("3: Call RequestDpcForIsr()\n");
      OsaConsoleGetStr(szBuf);
      nCmd = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      DemoLogf("Issuing Driver Debug Command %d\n", nCmd);
      OtxDrvDbgDump(g_hPhysDevs.hThorController0, nCmd);
    }
    break;
  case 'O':
    {
      OTX_INT32 ts; 
      OTX_UINT8 sqTimeSlotMask[32];


      // Connect pipe: Dsp -> Li 
      //DemoLogf("OtxDrvConnectPipe\n");
      result = OtxDrvDisconnectPipe(g_sqDtmfDialer[1].hPipe, OTX_PIPE_INPUT); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
      result = OtxDrvDisconnectPipe(g_sqDtmfDialer[2].hPipe, OTX_PIPE_INPUT); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

#if 0 // Arbitrary timeslot on the DSP highway
      result = OtxDrvConnectPipe(g_sqDtmfDialer[2].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[0], -1, NULL, 0); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
      result = OtxDrvConnectPipe(g_sqDtmfDialer[1].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[0], -1, NULL, 0); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
#else
      ts = 2;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_sqDtmfDialer[1].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[0], -1, sqTimeSlotMask, 32); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

      ts = 1;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_sqDtmfDialer[2].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[0], -1, sqTimeSlotMask, 32); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
#endif

      result = OtxDrvConnectLogicalDevice(g_sqDtmfDialer[1].hHandle, -1, g_sqDtmfDialer[1].hPipe,   -1); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
      result = OtxDrvConnectLogicalDevice(g_sqDtmfDialer[2].hHandle, -1, g_sqDtmfDialer[2].hPipe,   -1); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

    }
    break;
  case 'j':
    {
      char szBuf[12];
      OTX_INT32 nDsp;
      
      DemoLogf("Dump debug data. Which Dsp#?");
      OsaConsoleGetStr(szBuf);
      nDsp = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 0);
    }
    break;
  case 'J':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < 2; nDsp++) {
        DemoLogf("Dump debug data for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 0);
      }
    }
    break;
#endif

#if 1 // Only for debugging
 case 'p':
    {
      OTX_INT32 nDialer;
      
      DemoLogf("Disable/Enable DTMF dialler[0-3]\n");
      for (nDialer = 0; nDialer < 32; nDialer++) {
        if (nDialer%16 == 0) continue; // Skip 0 (sync), 16 (signalling), 32 (sync), 48 (signalling)
        result = OtxDrvDisable(g_sqDtmfDialer[nDialer].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDialer[nDialer].hHandle, result);
      }
      for (nDialer = 0; nDialer < 32; nDialer++) {
        if (nDialer%16 == 0) continue; // Skip 0 (sync), 16 (signalling), 32 (sync), 48 (signalling)
        result = OtxDrvEnable(g_sqDtmfDialer[nDialer].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDialer[nDialer].hHandle, result);
      }
      DemoLogf("Done.");
    }
    break;
 case 'P':
    {
      OTX_INT32 nDialer;
      static OTX_INT32 nLoopTimes = 1;
      OTX_INT32 i;
      char strBuf[12];
      
      DemoLogf("Reseting DTMF dialler[0-3]\n");
      DemoLogf("How many times? (%d)", nLoopTimes);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nLoopTimes = strtol(strBuf, NULL, 0);
      }

      for (i = 0; i < nLoopTimes; i++) {
        for (nDialer = 0; nDialer < 32; nDialer++) {
          if (nDialer%16 == 0) continue; // Skip 0 (sync), 16 (signalling), 32 (sync), 48 (signalling)
          result = OtxDrvDisable(g_sqDtmfDialer[nDialer].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDialer[nDialer].hHandle, result);
        }
        for (nDialer = 0; nDialer < 32; nDialer++) {
          if (nDialer%16 == 0) continue; // Skip 0 (sync), 16 (signalling), 32 (sync), 48 (signalling)
          result = OtxDrvEnable(g_sqDtmfDialer[nDialer].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDialer[nDialer].hHandle, result);
        }
      }
      DemoLogf("Done.");
    }
    break;
#endif


#if 1 // Only for debugging
  case 'b':
    {
      char szBuf[12];
      OTX_INT32 nDsp;
      
      DemoLogf("Init Bsp. Which Dsp#?");
      OsaConsoleGetStr(szBuf);
      nDsp = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);

    }
    break;
  case 'B':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < 2; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
    break;
#endif


  case 'f':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < 2; nLi++) {
        DemoLogf("Forcing resynchronization on Li #%d... ", nLi);
        result = OtxT1E1ForceResynch(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;


  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < 2; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;



  case 'v':
    {
      OTX_INT32 i;
      g_bHdlcVerbose = !g_bHdlcVerbose;
      DemoLogf("g_bHdlcVerbose is now %s.\n", g_bHdlcVerbose ? "TRUE": "FALSE");
      for (i = 0; i < 256; i++) {
        if (g_coHdlcRxFrames[i]) {
          DemoLogf("g_coHdlcRxFrames[%d] = %d (reset to 0) \n", i, g_coHdlcRxFrames[i]);
          g_coHdlcRxFrames[i] = 0;
          g_coHdlcBadRxFrames[i] = 0;
        }
      }
    }
    break;

  case 'V':
    {
      OTX_INT32 i;
      for (i = 0; i < 256; i++) {
        if (g_coHdlcRxFrames[i]) {
          DemoLogf("Tot[%.2d]=%.6d (Bad[%.2d]=%.6d)  ", i, g_coHdlcRxFrames[i], i, g_coHdlcBadRxFrames[i]);
          if (i && i % 2) {
            DemoLogf("\n");
          }
        }
      }
      DemoLogf("\n");
    }
    break;


  case 'a':
    {

      DemoLogf("Send AIS (All Ones) on Li #%d is ", nCurrentLi);
      result = OtxT1E1AlarmOn(g_hPhysDevs.hLi[nCurrentLi], OTX_T1E1_ALARM_AIS); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'A':
    {

      DemoLogf("Stop Sending AIS (All Ones) on Li #%d is ", nCurrentLi);
      result = OtxT1E1AlarmOff(g_hPhysDevs.hLi[nCurrentLi], OTX_T1E1_ALARM_AIS); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      DemoLogf("Done.\n");
    }
    break;


  case ';':
    {
      
      g_bAutoTransmit[nCurrentLi] = !g_bAutoTransmit[nCurrentLi];

      DemoLogf("Auto transmit of HDLC frames is now turned %s for Li #%d \n", g_bAutoTransmit[nCurrentLi] ? "ON" : "OFF", nCurrentLi);
      
      if (g_bAutoTransmit[nCurrentLi]) {
      	OsaThreadStartEx(DemoAutoTransmitHdlcFunction, (void *)&nCurrentLi);
      } 
            
    }
    break;


  case 't':
    {
      OTX_INT32 i;
      OTX_INT32 busyCount = 0;
#if !defined(_WIN32_WCE)
      clock_t startTime, stopTime;
#endif

      if (g_cbLiHdlcSenderFrame > OD_HDLC_MAX_TX_HDLC_FRAME_LEN) {
        DemoLogf("Error: Frame Length too long\n");
        break;
      }

      g_cbHdlcFrame = g_cbLiHdlcSenderFrame;
      for (i = 0; i < g_cbHdlcFrame; i++) {
        g_sqHdlcFrame[i] = (OTX_UINT8) i;
      }


      DemoLogf("Sending %d HDLC frame(s) (TS16) from Li #%d (len=%d) ...", g_coHdlcSenderFrames, nCurrentLi, g_cbLiHdlcSenderFrame);

      // Send HDLC Frame(s)
#if !defined(_WIN32_WCE)
      startTime = clock();
#endif
      for (i = 0; i < g_coHdlcSenderFrames; i++) {
        result = OtxHdlcSendData(g_sqLiHdlcSender[nCurrentLi].hHandle, g_sqHdlcFrame, g_cbHdlcFrame, DEMO_TASK_REF, 200); 
        if (g_bHdlcVerbose) {
          DemoCheckRc(g_sqLiHdlcSender[nCurrentLi].hHandle, result);
        } else {
          if (result == OTX_E_BUSY) {
            // Driver IO Request buffer is full. We need to retransmit this frame.
            i--;
            busyCount++;
          } else {
            DemoCheckRc(g_sqLiHdlcSender[nCurrentLi].hHandle, result);
          }
        }
      }
#if !defined(_WIN32_WCE)
      stopTime = clock();
      DemoLogf("Done (BusyCount=%d, time=%2.1f seconds).\n", busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
#endif
    }
    break;


  case 'T':
    {
      OTX_INT32 i;
      OTX_INT32 busyCount = 0;
#if !defined(_WIN32_WCE)
      clock_t startTime, stopTime;
#endif
#if 0 // For Driver Timing Profiling      
      OTX_DATETIME nStartTime;
      OTX_DATETIME nStopTime;
#endif

      if (g_cbDspHdlcSenderFrame > OD_HDLC_MAX_TX_HDLC_FRAME_LEN) {
        DemoLogf("Error: Frame Length too long\n");
        break;
      }

      g_cbHdlcFrame = g_cbDspHdlcSenderFrame;
      for (i = 0; i < g_cbHdlcFrame; i++) {
        g_sqHdlcFrame[i] = (OTX_UINT8) i;
      }

      DemoLogf("Sending %d HDLC frame(s) (TS1) from DSP #%d (len=%d)...", g_coHdlcSenderFrames, nCurrentHdlcSender, g_cbDspHdlcSenderFrame);

      // Send HDLC Frame(s)
#if !defined(_WIN32_WCE)
      startTime = clock();
#endif
#if 0 // For Driver Timing Profiling      
      QueryPerformanceCounter((LARGE_INTEGER *)&nStartTime);
#endif

      for (i = 0; i < g_coHdlcSenderFrames; i++) {
        result = OtxHdlcSendData(g_sqDspHdlcSender[nCurrentHdlcSender].hHandle, g_sqHdlcFrame, g_cbHdlcFrame, DEMO_TASK_REF, 200); 
        if (g_bHdlcVerbose) {
          DemoCheckRc(g_sqLiHdlcSender[nCurrentLi].hHandle, result);
        } else {
          if (result == OTX_E_BUSY) {
            // Driver IO Request buffer is full. We need to retransmit this frame.
            i--;
            busyCount++;
          } else {
            DemoCheckRc(g_sqDspHdlcSender[nCurrentHdlcSender].hHandle, result);
          }
        }
      }
#if 0 // For Driver Timing Profiling      
      QueryPerformanceCounter((LARGE_INTEGER *)&nStopTime);
#endif
#if !defined(_WIN32_WCE)
      stopTime = clock();
      DemoLogf("Done (BusyCount=%d, time=%2.1f seconds).\n", busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
#endif

#if 0 // For Driver Timing Profiling      
      {
        OTX_DATETIME nTimeDiff;
        DemoLogf("StartTime: "); DemoPrintTimeStamp(nStartTime); DemoLogf("\n");
        DemoLogf("StopTime : "); DemoPrintTimeStamp(nStopTime);  DemoLogf("\n");
        nTimeDiff.nUint64 = nStopTime.nUint64 - nStartTime.nUint64;
        DemoLogf("TimeDiff : "); DemoPrintTimeStamp(nTimeDiff); DemoLogf("\n");
        
        DemoLogf("TimeStamps for Driver Device %d (see windbg)\n", 0);
        OtxDrvDbgDump(OTX_DRIVER_DEVICE_HANDLE, 100);
        
        DemoLogf("TimeStamps for HDLC Sender Logical device %d (see windbg)\n", OTX_LDEVICE_HDLC_SENDER);
        OtxDrvDbgDump(g_sqDspHdlcSender[nCurrentHdlcSender].hHandle, 100);
        
        DemoLogf("TimeStamps for DSP device %d (see windbg)\n", OTX_DEVICE_DSP);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[0], 100);
        
        DemoLogf("TimeStamps for Board device %d (see windbg)\n", OTX_DEVICE_THOR_PCMCIA);
        OtxDrvDbgDump(g_hPhysDevs.hThorController0, 100);
      }
#endif

    }
    break;


  case 'd':
    {
      // Send DTMF from Dtmf Dialer

      switch (g_demoEnv.eProgramModule) {
      case OTX_DSP_SPM_OTXHDLC:
        DemoLogf("DTMF dialing is not supported for the OTXHDLC SPM module.\n");
        break;    
      case OTX_DSP_SPM_OTXSPM1:
      case OTX_DSP_SPM_OTXSPM2:
        {
          char g_sqDigits[] = "0123456789ABCD*#";
          //char g_sqDigits[] = "0";
          OTX_INT32 nDigits;
          OTX_DATETIME nTimeStamp;
          OTX_INT32 nDialer = nCurrentLi;
          
          nDigits = strlen(g_sqDigits);
          OtxDrvGetCurrentTime(&nTimeStamp);
          DemoLogf("%u%u DtmfDialer[%d] Dialling digits %s.\n", nTimeStamp.sUint32.nHigh, nTimeStamp.sUint32.nLow/10000, nDialer, g_sqDigits);
          
          // Start Dialing DTMF number, Non-Blocking call, will receive Event when completed
          result = OtxToneDtmfDial(g_sqDtmfDialer[nDialer].hHandle, (OTX_UINT8 *)g_sqDigits, nDigits, DEMO_TASK_REF, OTX_TIME_INFINITY); 
          if (OTX_RESULT_CODE(result) == OTX_DTMF_DIALER_RES_BUSY) {
            DemoLogf("Dialer busy.\n");
          } else {
            DemoCheckRc(g_sqDtmfDialer[nDialer].hHandle, result);
          }
          
        }
        break;
      case OTX_DSP_SPM_OTXDTTX:
        {
          // Use the DTMF Dialer thread
          char g_sqDigits[] = "1";
          OTX_INT32 nDialer = nCurrentDialer;
          
          if (g_sqDtmfDialer[nDialer].bActive) {
            DemoLogf("DTMF Dialer Thread busy for Dialer#%d. Aborting this operation...\n", nDialer);
            break;
          }
          
          
          strcpy(g_sqDtmfDialer[nDialer].sqDigits, g_sqDigits);
          g_sqDtmfDialer[nDialer].sqDigits[0] = '0' + (nDialer % 10);
          g_sqDtmfDialer[nDialer].nRepeatTimes = 1; 

          DemoLogf("Starting DTMF Dialer Thread#%d\n", nDialer);
          g_sqDtmfDialer[nDialer].bActive = OTX_TRUE;
          
        }
        break;
      default:
        DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
        OsaConsoleExit();
        OtxDrvReleaseLib();
        exit(1);
      }

    }
    break;

  case 'D':
    {
      char g_sqDigits[] = "1234512345";
      //char g_sqDigits[] = "0";
      OTX_INT32 nDigits;
      OTX_DATETIME nTimeStamp;
      OTX_INT32 nDialer = nCurrentLi;
      OTX_INT32 i;
      
      nDigits = strlen(g_sqDigits);
      OtxDrvGetCurrentTime(&nTimeStamp);
      DemoLogf("%u%u DtmfDialer[%d] Dialling digits %s.\n", nTimeStamp.sUint32.nHigh, nTimeStamp.sUint32.nLow/10000, nDialer, g_sqDigits);
      
      // Start Dialing DTMF number, Non-Blocking call, will receive Event when completed
      result = OtxToneDtmfDial(g_sqDtmfDialer[nDialer].hHandle, (OTX_UINT8 *)g_sqDigits, nDigits, DEMO_TASK_REF, OTX_TIME_INFINITY); 
      if (OTX_RESULT_CODE(result) == OTX_DTMF_DIALER_RES_BUSY) {
        DemoLogf("Dialer busy.\n");
      } else {
        DemoCheckRc(g_sqDtmfDialer[nDialer].hHandle, result);
      }

      for (i = 0; i < 500; i++) {
        OTX_BOOL  bSilence = 0;      

        result = OtxToneSilenceDetectorStatus(g_sqSilenceDetector[!nDialer].hHandle, &bSilence); 
        DemoCheckRc(g_sqSilenceDetector[!nDialer].hHandle, result);

        // @parm OTX_TRUE if silence, OTX_FALSE if tone detected
        if (bSilence) {
          DemoLogf("s");
        } else {
          DemoLogf(".");
        }
        OsaThreadSleep(5);

      }
      DemoLogf("\n");
      
    }
    break;

  case ')':
    {
      // Turn on Constant Value on Dialer
      OTX_INT32 i;
      OtxDeviceStateE  eDevState;


      if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXDTTX) {
        
        DemoLogf("Turn on Constant Value on Dialer #%1-15 and 17-31. (ConstVal=TS]");
        
        for (i = 0; i < 32; i++) {
          if (i == 0) continue; // Sync channel
          if (i == 16) continue; // HDLC channel
          
          result = OtxToneConstValGeneratorInit(g_sqConstValGenerator[i].hHandle, OTX_DATA_FORMAT_UINT8, (OTX_UINT16)i); DemoCheckRc(g_sqConstValGenerator[i].hHandle, result);
          if (result != OTX_S_OK) {
            break;
          }
          
          
          result = OtxDrvGetState(g_sqConstValGenerator[i].hHandle, &eDevState); DemoCheckRc(g_sqConstValGenerator[i].hHandle, result);
          if (result != OTX_S_OK) {
            break;
          }
          
          
          DemoLogf("Enabling ConstValGenerator[%d] device.\n", i);
          if (eDevState != OTX_DEVICE_WORKING) {
            result = OtxDrvEnable(g_sqConstValGenerator[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqConstValGenerator[i].hHandle, result);
          } else {
            DemoLogf("Device already enabled. No need to enable it again\n");
          }

        }

      } else {
        DemoLogf("This function is not implemented for this SPM module\n");
      }

    }
    break;


  case 'c':
    {
      // Turn on Constant Value on Dialer
      char szBuf[12];
      static OTX_UINT16 nConstVal = 0x55;
      OtxDeviceStateE  eDevState;


      if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXDTTX) {

        DemoLogf("Turn on Constant Value on Dialer #%d. Enter value (0x%.2X)?", nCurrentDialer, nConstVal);
        OsaConsoleGetStr(szBuf);
        if (*szBuf) {
          nConstVal = (OTX_UINT16)strtol(szBuf, NULL, 0);
        }

        result = OtxToneConstValGeneratorInit(g_sqConstValGenerator[nCurrentDialer].hHandle, OTX_DATA_FORMAT_UINT8, nConstVal); DemoCheckRc(g_sqConstValGenerator[nCurrentDialer].hHandle, result);
        if (result != OTX_S_OK) {
          break;
        }


        result = OtxDrvGetState(g_sqConstValGenerator[nCurrentDialer].hHandle, &eDevState); DemoCheckRc(g_sqConstValGenerator[nCurrentDialer].hHandle, result);
        if (result != OTX_S_OK) {
          break;
        }


        DemoLogf("Enabling ConstValGenerator device.\n");
        if (eDevState != OTX_DEVICE_WORKING) {
          result = OtxDrvEnable(g_sqConstValGenerator[nCurrentDialer].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqConstValGenerator[nCurrentDialer].hHandle, result);
        } else {
          DemoLogf("Device already enabled. No need to enable it again\n");
        }

      } else {
        DemoLogf("This function is not implemented for this SPM module\n");
      }

    }
    break;

  case 'C':
    {
      // Turn off Constant Value on current dialer

      if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXDTTX) {

        DemoLogf("Turn off Constant Value on Dialer #%d \n", nCurrentDialer);

        result = OtxDrvDisable(g_sqConstValGenerator[nCurrentDialer].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqConstValGenerator[nCurrentDialer].hHandle, result);

      } else {
        DemoLogf("This function is not implemented for this SPM module\n");
      }

    }
    break;

  case 's':
    {
      // Get sample value from a Dialer
      OTX_UINT16 nSampleVal;

      if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXDTTX) {

        DemoLogf("ValSampler[%d]: ", nCurrentDialer);

        result = OtxToneValSamplerReadData(g_sqValSampler[nCurrentDialer].hHandle, &nSampleVal); DemoCheckRc(g_sqValSampler[nCurrentDialer].hHandle, result);

        DemoLogf("0x%.2X \n", nSampleVal);

      } else {
        DemoLogf("This function is not implemented for this SPM module\n");
      }

    }
    break;


  case 'S':
    {
      // Get sample value from a Dialer
      OTX_UINT16 nSampleVal;
      OTX_INT32 i;

      if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXDTTX) {

        DemoLogf("ValSampler for Li0: \n");
        for (i = 0; i < 32; i++) {
          if ((i != 0) && ((i%8) == 0)) DemoLogf("\n");

          if ((i == 0) || (i == 16)) {  // Sync channel or HDLC channel
            DemoLogf("[%.2d]: xx ", i);
            continue; 
          }
          
          result = OtxToneValSamplerReadData(g_sqValSampler[i].hHandle, &nSampleVal); DemoCheckRc(g_sqValSampler[i].hHandle, result);
          
          DemoLogf("[%.2d]: %.2X ", i, nSampleVal);
        }

        DemoLogf("\n\n");
 
      } else {
        DemoLogf("This function is not implemented for this SPM module\n");
      }

    }
    break;


#if 1 // for debugging
  case 'u':
    {
      // Continously send a short DTMF string from all Dtmf Dialers
      char g_sqDigits[] = "5";
      OTX_INT32 nDialer;
      
      if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXDTTX) {
        for (nDialer = 0; nDialer < 64; nDialer++) {
          
          if (nDialer%16 == 0) continue; // Skip 0 (sync), 16 (signalling), 32 (sync), 48 (signalling)
          
          if (g_sqDtmfDialer[nDialer].bActive) {
            DemoLogf("DTMF Dialer Thread busy for Dialer#%d. Aborting this operation...\n", nDialer);
            continue;
          }
          
          strcpy(g_sqDtmfDialer[nDialer].sqDigits, g_sqDigits);
          g_sqDtmfDialer[nDialer].nRepeatTimes = 100;
          
          DemoLogf("Starting DTMF Dialer Thread#%d (nRepeatTimes=%d)\n", nDialer, g_sqDtmfDialer[nDialer].nRepeatTimes);
          g_sqDtmfDialer[nDialer].bActive = OTX_TRUE;

        }
        
      } else {
        DemoLogf("This function is not implemented for this SPM module\n");
      }
      
    }
    break;
  case 'U':
    {
      // Print statistics from Dtmf Dialers
      OTX_INT32 nDialer;

      DemoLogf("Statistics from DTMF Dialers");
      
      for (nDialer = 0; nDialer < 64; nDialer++) {
        
        if (nDialer%4 == 0) DemoLogf("\n");

        if (nDialer%16 == 0) continue; // Skip 0 (sync), 16 (signalling), 32 (sync), 48 (signalling)

        DemoLogf("[%.2d]: %.8d  ", nDialer, g_sqDtmfDialer[nDialer].coDtmfCompleted);

      }
      DemoLogf("\n");
      
    }
    break;
#endif

  case 'z':
    {
      OtxT1E1TermModeE eTermMode;
      DemoLogf("Set Line Termination mode:\n");
      DemoLogf("0 = OTX_T1E1_TERM_120_OHM\n");
      DemoLogf("1 = OTX_T1E1_TERM_100_OHM\n");
      DemoLogf("2 = OTX_T1E1_TERM_75_OHM\n");
      DemoLogf("3 = OTX_T1E1_TERM_HIGHZ\n");
      switch (OsaConsoleGetCh() - '0') {
      case 0:
        eTermMode = OTX_T1E1_TERM_120_OHM;
        break;
      case 1:
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      case 2:
        eTermMode = OTX_T1E1_TERM_75_OHM;
        break;
      case 3:
        eTermMode = OTX_T1E1_TERM_HIGHZ;
        break;
      default:
        DemoLogf("Invalid entry. Setting Line Termination to OTX_T1E1_TERM_100_OHM\n");
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      }

      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[nCurrentLi], eTermMode); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    }
    break;

  case 'n':
    {
      char strBuf[12];

      DemoLogf("Set Frame Length for Li%d HDLC Sender (%d):\n", nCurrentLi, g_cbLiHdlcSenderFrame);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_cbLiHdlcSenderFrame = (OTX_INT32) strtol(strBuf, NULL, 0);
        DemoLogf("Frame Length is now %d octets\n", g_cbLiHdlcSenderFrame);
      } else {
        DemoLogf("Frame Length remains unchanged\n");
      }
    }
    break;

  case 'N':
    {
      char strBuf[12];

      DemoLogf("Set Frame Length for DSP%d HDLC Sender (%d):\n", nCurrentLi, g_cbDspHdlcSenderFrame);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_cbDspHdlcSenderFrame = (OTX_INT32) strtol(strBuf, NULL, 0);
        DemoLogf("Frame Length is now %d octets\n", g_cbDspHdlcSenderFrame);
      } else {
        DemoLogf("Frame Length remains unchanged\n");
      }
    }
    break;

  case '+':
    {
      char strBuf[12];

      DemoLogf("Set the number of consecutive frame to sent when 't' or 'T' are pressed (%d)\n", g_coHdlcSenderFrames);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_coHdlcSenderFrames = (OTX_INT32) strtol(strBuf, NULL, 0);
        DemoLogf("Consecutive frame count is now %d\n", g_coHdlcSenderFrames);
      } else {
        DemoLogf("Consecutive frame count remains unchanged\n");
      }
    }
    break;

  case '0':
    // Select Li/Dsp 0 as current
    nCurrentLi = 0;
    DemoLogf("Li/Dsp#0 Selected ....\n");
    break;

  case '1':
    // Select Li/Dsp 1 as current
    nCurrentLi = 1;
    DemoLogf("Li/Dsp#1 Selected ....\n");
    break;


  case 'm':
    {
      switch (g_demoEnv.eProgramModule) {
      case OTX_DSP_SPM_OTXSPM1:
      case OTX_DSP_SPM_OTXSPM2:
        DemoLogf("Key not implemented for this module\n");
        break;
      case OTX_DSP_SPM_OTXDTTX:
        // Select current nCurrentDialer
        nCurrentDialer++;
        if (nCurrentDialer > 63) {
          // Wrap (and skip TS 0)
          nCurrentDialer = 1;
        } else if (nCurrentDialer%16 == 0) {
          // TS16 is signalling channel. Skip one up
          nCurrentDialer++;
        }
        
        DemoLogf("Dialer#%d (Li%d, TS%d) Selected \n", nCurrentDialer, nCurrentDialer/32, nCurrentDialer%32);
        break;
      case OTX_DSP_SPM_OTXHDLC:
        // Select current nCurrentHdlcSender
        nCurrentHdlcSender++;
        if (nCurrentHdlcSender > 7) {
          // Wrap 
          nCurrentHdlcSender = 0;
        }
        
        DemoLogf("DSP HDLC Sender#%d (DSP%d) Selected \n", nCurrentHdlcSender, nCurrentHdlcSender/4);
        break;    
      default:
        DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
        OsaConsoleExit();
        OtxDrvReleaseLib();
        exit(1);
      }

    }
    break;
    
    
  case 'M':
    {
      switch (g_demoEnv.eProgramModule) {
      case OTX_DSP_SPM_OTXSPM1:
      case OTX_DSP_SPM_OTXSPM2:
        DemoLogf("Key not implemented for this module\n");
        break;
      case OTX_DSP_SPM_OTXDTTX:
        // Select current nCurrentDialer
        nCurrentDialer--;
        if (nCurrentDialer < 1) {
          // Wrap (and skip TS 0)
          nCurrentDialer = 63;
        } else if (nCurrentDialer%16 == 0) {
          // TS16 is signalling channel. Skip one down
          nCurrentDialer--;
        }
        
        
        DemoLogf("Dialer#%d (Li%d, TS%d) Selected \n", nCurrentDialer, nCurrentDialer/32, nCurrentDialer%32);
        break;
      case OTX_DSP_SPM_OTXHDLC:
        // Select current nCurrentHdlcSender
        nCurrentHdlcSender--;
        if (nCurrentHdlcSender < 0) {
          // Wrap
          nCurrentHdlcSender = 7;
        }
        
        DemoLogf("DSP HDLC Sender#%d (DSP%d) Selected \n", nCurrentHdlcSender, nCurrentHdlcSender/4);
        break;    
      default:
        DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
        OsaConsoleExit();
        OtxDrvReleaseLib();
        exit(1);
      }
    }
    break;
    
        
  case 'r':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'w':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

  case '(':
    {
      OsaThreadStart(DemoTimeStampTest);
    }
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}                      


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}


int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  printf("T2MDEMO, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Thor-2-PCMCIA Example program.\n");

  OsaMutexInit(&g_analyzeCritSec);
  OsaMutexInit(&g_demoLogfCritSec);

  // Initialize global variables
  memset(g_sqDtmfDialer, 0, sizeof(g_sqDtmfDialer));

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }
  atexit(DemoAtExit);

  if (g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXDTTX) {
    if (g_demoEnv.eT1E1Mode != OTX_T1E1_MODE_E1) {
      DemoLogf("Error: This demo program only supports E1-mode for the OTXDTTX module\n.");
      OsaConsoleExit();
      exit(-1);
    }
  }


  {
    char *pszClockSource;

    switch (g_demoEnv.eClkSource) {
    case OTX_CLOCK_SOURCE_INTERNAL:
      pszClockSource = "Internal";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_0:
      pszClockSource = "Line Interface 0";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_1:
      pszClockSource = "Line Interface 1";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_2:
      pszClockSource = "External";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_3:
      pszClockSource = "Split Clock, both Line Interface are synchronized to its T1E1 span";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_4:
      pszClockSource = "Split Clock, Li0 synchronized to its T1E1 span. Li1 internal";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_5:
      pszClockSource = "Split Clock, Li1 synchronized to its T1E1 span. Li0 internal";
      break;
    default:
      pszClockSource = "Unknown";
      break;
    }

    DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
    DemoLogf("Clock Source: %s\n", pszClockSource);
  }

  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib();
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }


#ifndef DEMO_USE_TIMESTAMP_FROM_1601
  {
    OTX_UINT32 nTimeStampTickFrequency;
    OTX_BOOL bTimeStampFrom1601;

    result = OtxDrvGetAttributeValueUint32(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_TIMESTAMP_TICK_FREQUENCY, &nTimeStampTickFrequency); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    DemoLogf("Default TimeStampTickFrequency = %d \n", nTimeStampTickFrequency);
    g_nFrequency = nTimeStampTickFrequency;

    result = OtxDrvGetAttributeValueBool(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_TIMESTAMP_FROM_1601, &bTimeStampFrom1601); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    DemoLogf("Using TimeStamps from 1601 = %s \n", bTimeStampFrom1601 ? "Yes" : "No");

    DemoLogf("Setting Using TimeStamps from 1601 to OTX_FALSE\n");
    bTimeStampFrom1601 = OTX_FALSE;
    result = OtxDrvSetAttributeValueBool(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_TIMESTAMP_FROM_1601, bTimeStampFrom1601); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_TIMESTAMP_FROM_1601, &bTimeStampFrom1601); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    DemoLogf("Using TimeStamps from 1601 = %s \n", bTimeStampFrom1601 ? "Yes" : "No");

    result = OtxDrvGetAttributeValueUint32(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_TIMESTAMP_TICK_FREQUENCY, &nTimeStampTickFrequency); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    DemoLogf("New TimeStampTickFrequency = %d \n", nTimeStampTickFrequency);

#ifdef WIN32
    {
      LARGE_INTEGER nFrequency;
      QueryPerformanceFrequency(&nFrequency);
      DemoLogf("QueryPerformanceFrequency() = %d (should be the same as new TimeStampTickFrequency)\n", nFrequency);
    }
#endif

  }
#endif // DEMO_USE_TIMESTAMP_FROM_1601

  // Query driver revision
  {
    char szVersionInfo[255];
    result = OtxDrvGetAttributeValueSz(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

    DemoLogf("Driver version %s\n", szVersionInfo);
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCMCIA board
  DemoLogf("DemoCreatePhysicalDevices\n");
  DemoCreatePhysicalDevices();

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnablePhysicalDevices\n");
  DemoEnablePhysicalDevices();

  // Set clock source for the board
  if (g_hPhysDevs.hVidarAsmController0) {
    // Thor-2-PCI
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, g_demoEnv.eClkSource); DemoCheckRc(g_hPhysDevs.hThorController0, result);
#if 1  // Temporary
    {
      OTX_INT32 nDsp;
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
#endif
  } else {
    // Thor-2-PCMCIA
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_NA, g_demoEnv.eClkSource); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  }

#if 0 // For testing
  {
    OTX_INT32 i;
    
    for (i = 0; i < 1000; i++) {
      // Create and close the Logical devices needed by this application
      DemoLogf("Creating and closing logical devices [%d]...\n", i);
      DemoCreateLogicalDevices();
      DemoCloseLogicalDevices();
    }
  }
#endif

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("DemoCreateLogicalDevices\n");
  DemoCreateLogicalDevices();

#if 0 // For testing 
  {
    DemoTestDataConverter();
  }
#endif


  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("DemoConnectDevices\n");
  DemoConnectDevices();

  // Initialize the logical devices. A logical device must be create and connected before it can be initialized.
  DemoLogf("DemoInitializeLogicalDevices\n");
  DemoInitializeLogicalDevices();

  // Enable logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnableLogicalDevices\n");
  DemoEnableLogicalDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *)malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *)malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  // Read User String from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdUserString;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_USER_STRING, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdUserString = (char *)malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_USER_STRING, pszBrdUserString, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("User String: %s\n", pszBrdUserString);
    free(pszBrdUserString);
  }

  // Read Bus Address from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    OTX_UINT32      nBusAddress;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUS_ADDRESS, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_UINT32);
    assert(bReadOnly == OTX_TRUE);
    assert(cbAttribValue == sizeof(OTX_UINT32));
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUS_ADDRESS, &nBusAddress); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Bus Address: 0x%X\n", nBusAddress);
  }


  // Read Pnp Id from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    OTX_UINT32      cbwcszPnpId;
    wchar_t *pwcszPnpId;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PNP_ID, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_OCTETS); // We read the PnP Id as a string of octets from the OTX driver, but the value is actually a wide-character string.
    assert(bReadOnly == OTX_TRUE);
    assert((cbAttribValue % 2) == 0); 
    if (cbAttribValue) {
      pwcszPnpId = (wchar_t *)malloc(cbAttribValue);
      result = OtxDrvGetAttributeValue(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PNP_ID, (OTX_PVOID)pwcszPnpId, cbAttribValue, &cbwcszPnpId); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      DemoLogf("PnP Id: %S\n", pwcszPnpId);
      free(pwcszPnpId);
    } else {
      DemoLogf("PnP Id: None\n");
    }
  }


  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_8_PCI:
    // Query PCI Device and PCI Function Number for this PCI board
    {
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      nPciDeviceNumber;
      OTX_UINT32      nPciFunctionNumber;
      
      result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PCI_DEVICE_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      assert(eAttribType == OTX_ATTRIB_UINT32);
      assert(bReadOnly == OTX_TRUE);
      assert(cbAttribValue == sizeof(OTX_UINT32));
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PCI_DEVICE_NUMBER, &nPciDeviceNumber); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("PCI Device Number : %.4X\n", nPciDeviceNumber);

      result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PCI_FUNCTION_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      assert(eAttribType == OTX_ATTRIB_UINT32);
      assert(bReadOnly == OTX_TRUE);
      assert(cbAttribValue == sizeof(OTX_UINT32));
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PCI_FUNCTION_NUMBER, &nPciFunctionNumber); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("PCI Function Number: %.4X\n", nPciFunctionNumber);

    }
    break;
  case OTX_DEVICE_THOR_PCMCIA:
    // Query PCMCIA Socket Number for this PCMCIA board
    {
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      nPcmciaSocketNumber;
      
      result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PCMCIA_SOCKET_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      assert(eAttribType == OTX_ATTRIB_UINT32);
      assert(bReadOnly == OTX_TRUE);
      assert(cbAttribValue == sizeof(OTX_UINT32));
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PCMCIA_SOCKET_NUMBER, &nPcmciaSocketNumber); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      DemoLogf("PCMCIA Socket Number: %d\n", nPcmciaSocketNumber);
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  // Create all DTMF Dialer threads 
  {
    OTX_INT32 nDialer;
    
    for (nDialer = 0; nDialer <= DEMO_DIALER_MAX; nDialer++) {
      if (nDialer%16 == 0) continue; // Skip 0 (sync), 16 (signalling), 32 (sync), 48 (signalling)
      g_sqDtmfDialer[nDialer].nDialer = nDialer; 

    	OsaThreadStartEx(DemoDtmfDialer, (void *)&g_sqDtmfDialer[nDialer].nDialer);
    }
    
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300); // Only so the following message will show on the screen

#if 0
  // Enable DSP Serial port clocks (so data can flow on the DSP highways)
  result = OtxDrvSetAttributeValueBool(
    g_hPhysDevs.hThorController0, 
    OTX_ATTR_BRD_DSP_PORTS_DISABLED, 
    OTX_FALSE
    ); 
  DemoCheckRc(g_hPhysDevs.hThorController0, result);  
#endif

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }


  // Clean up
#if 0
  {
    OTX_INT32 nDialer;
    OTX_BOOL bAllThreadsSuspended; 
    
    for (nDialer = 0; nDialer <= DEMO_DIALER_MAX; nDialer++) {
      if (nDialer%16 == 0) continue; // Skip 0 (sync), 16 (signalling), 32 (sync), 48 (signalling)
      g_sqDtmfDialer[nDialer].bAbort = OTX_TRUE; 
    }
    DemoLogf("Waiting for all DTMF Dialler Threads to goto Suspend state\n");
    do {
      bAllThreadsSuspended = OTX_TRUE; 
      for (nDialer = 0; nDialer <= DEMO_DIALER_MAX; nDialer++) {
        if (nDialer%16 == 0) continue; // Skip 0 (sync), 16 (signalling), 32 (sync), 48 (signalling)
        if (!DemoIsThreadSuspended(g_sqDtmfDialer[nDialer].hThread)) {
          bAllThreadsSuspended = OTX_FALSE;
        }
      }
    } while (!bAllThreadsSuspended);
    DemoLogf("All DTMF Dialler Threads are now in Suspend state. Exiting...\n");

  }
#endif

  OsaConsoleExit();
  OtxDrvReleaseLib();

  return 0;
}

#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  static char exename[_MAX_PATH * sizeof(TCHAR)] = "";
  static char *cmdParams[] = { exename, "/TM", "/P1", NULL }; 
  char *pc;

  GetModuleFileName(NULL, (LPTSTR) cmdParams[0], _MAX_PATH);
  TCHARTOSTR((LPTSTR) cmdParams[0], cmdParams[0]);
  strcpy(g_demoEnv.cwd, cmdParams[0]);
  pc = strrchr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
